# Most of the tests based on
# https://github.com/cta-observatory/cta-lstchain/blob/master/lstchain/scripts/tests/test_lstchain_scripts.py

import pytest
import subprocess
import pkg_resources


def find_entry_points(package_name):
    """from: https://stackoverflow.com/a/47383763/3838691"""
    entrypoints = [
        ep.name
        for ep in pkg_resources.iter_entry_points("console_scripts")
        if ep.module_name.startswith(package_name)
    ]
    return entrypoints


ALL_SCRIPTS = find_entry_points("lstmcpipe")


def run_script(*args):
    result = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding='utf-8')

    if result.returncode != 0:
        raise ValueError(
            f"Running {args[0]} failed with return code {result.returncode}"
            f", output: \n {result.stdout}"
        )


@pytest.mark.parametrize("script", ALL_SCRIPTS)
def test_all_help(script):
    """Test for all scripts if at least the help works."""
    run_script(script, "--help")



#!/usr/bin/env python

from ruamel.yaml import YAML
import pytest
from pathlib import Path


def test_save_log_to_file():
    from ..workflow_management import save_log_to_file
    outfile_yml = Path('dummy_log.yml')

    dummy_log = {
        'dummy_jobid': 'sbatch --parsable --wrap="sleep 10"'
    }
    save_log_to_file(dummy_log, outfile_yml)

    assert outfile_yml.exists()
    with open(outfile_yml) as f:
        log = YAML().load(f)

    assert "NoKEY" in log.keys()
    assert isinstance(log["NoKEY"], dict)
    assert "dummy_jobid" in log["NoKEY"].keys()
    assert dummy_log["dummy_jobid"] == log["NoKEY"]["dummy_jobid"]


@pytest.fixture()
def create_fake_dl1_structure(tmp_path):
    fake_dl1_dir = tmp_path.joinpath("DL1", "{}")

    gamma_off = ["off0.0deg", "off0.4deg"]
    dataset = ["training", "testing"]

    for particle in ["electron", "gamma-diffuse", "proton", "gamma"]:
        if particle == "gamma":
            for off in gamma_off:
                _particle = particle + "_" + off
                _subdir = Path(str(fake_dl1_dir).format(particle), off)
                _subdir.mkdir(parents=True)
                for i_set in dataset:
                    with open(_subdir.joinpath(f"dl1_{_particle}_dummy_{i_set}.h5"), "w"):
                        pass
        else:
            _subdir = Path(str(fake_dl1_dir).format(particle))
            _subdir.mkdir(parents=True)
            for i_set in dataset:
                with open(_subdir.joinpath(f"dl1_{particle}_dummy_{i_set}.h5"), "w"):
                    pass

    return fake_dl1_dir.as_posix()



from pathlib import Path
from lstmcpipe import prod_logs


def test_create_log_files():
    from ..lstmcpipe_tree_path import create_log_files
    logs, scancel_file, log_dir = create_log_files('dummy_prodID')

    assert "log_file" in logs.keys()
    assert "debug_file" in logs.keys()

    assert isinstance(logs["log_file"], Path)
    assert isinstance(logs["debug_file"], Path)
    assert isinstance(scancel_file, Path)

    assert scancel_file.exists()

    scancel_file.unlink()
    log_dir.rmdir()


def test_update_scancel_file():
    from ..lstmcpipe_tree_path import create_log_files, update_scancel_file
    _, scancel_file, log_dir = create_log_files("dummy_prodID")

    update_scancel_file(scancel_file, "1234")
    with open(scancel_file) as f:
        lines = f.readlines()
    assert lines == ["scancel 1234"]

    update_scancel_file(scancel_file, "5678")
    with open(scancel_file) as f:
        lines = f.readlines()
    assert lines == ["scancel 1234,5678"]

    scancel_file.unlink()
    log_dir.rmdir()


def test_backup_log():
    from ..lstmcpipe_tree_path import backup_log
    dummy_file = Path('./test_file.txt')
    dummy_file.touch()

    saved_file_0 = Path('./BACKUP_00_test_file.txt')
    saved_file_1 = Path('./BACKUP_01_test_file.txt')

    backup_log(dummy_file)
    assert saved_file_0.exists()
    backup_log(dummy_file)
    assert saved_file_1.exists()

    dummy_file.unlink()
    saved_file_0.unlink()
    saved_file_1.unlink()


def test_create_log_dir():
    from ..lstmcpipe_tree_path import create_log_dir
    log_dirname = create_log_dir("test_prodID")
    assert log_dirname.is_dir()
    assert log_dirname.name == prod_logs.joinpath("logs_test_prodID").name
    log_dirname.rmdir()



"""
Procedure adapted from pyirf v0.4 example one (used with eventdisplay)
"""
import logging
import operator

from pathlib import Path
import numpy as np
from astropy import table
import astropy.units as u
from astropy.io import fits

from pyirf.binning import (
    create_bins_per_decade,
    add_overflow_bins,
    create_histogram_table,
)
from pyirf.cuts import calculate_percentile_cut, evaluate_binned_cut
from pyirf.sensitivity import calculate_sensitivity, estimate_background
from pyirf.utils import calculate_theta, calculate_source_fov_offset
from pyirf.benchmarks import energy_bias_resolution, angular_resolution
from pyirf.benchmarks.energy_bias_resolution import energy_resolution_absolute_68

from pyirf.spectral import (
    calculate_event_weights,
    PowerLaw,
    CRAB_HEGRA,
    IRFDOC_PROTON_SPECTRUM,
    IRFDOC_ELECTRON_SPECTRUM,
)
from pyirf.cut_optimization import optimize_gh_cut

from pyirf.irf import (
    effective_area_per_energy,
    energy_dispersion,
    psf_table,
    background_2d,
)

from pyirf.io import (
    create_aeff2d_hdu,
    create_psf_table_hdu,
    create_energy_dispersion_hdu,
    create_rad_max_hdu,
    create_background_2d_hdu,
)

from lstchain.io.io import read_mc_dl2_to_QTable
import argparse


T_OBS = 50 * u.hour

# scaling between on and off region.
# Make off region 10 times larger than on region for better
# background statistics
ALPHA = 0.1

# Radius to use for calculating bg rate
MAX_BG_RADIUS = 1 * u.deg
MAX_GH_CUT_EFFICIENCY = 0.9
GH_CUT_EFFICIENCY_STEP = 0.01

# gh cut used for first calculation of the binned theta cuts
INITIAL_GH_CUT_EFFICENCY = 0.4

MIN_THETA_CUT = 0.1 * u.deg
MAX_THETA_CUT = 0.5 * u.deg

MIN_ENERGY = 20.0 * u.GeV
MAX_ENERGY = 20.05 * u.TeV

N_BIN_PER_DECADE = 5

filters = {
    'intensity': [20, np.inf],
    # 'leakage_intensity_width_2': [0, 0.2],
}



def determine_source_position(gamma_events):
    if len(set(gamma_events['true_alt'].value)) == 1:
        source_alt = gamma_events['true_alt'][0]
    else:
        raise ValueError("The gamma source position is not unique, one should be provided with --source-alt")

    if len(set(gamma_events['true_az'].value)) == 1:
        source_az = gamma_events['true_az'][0]
    else:
        raise ValueError("The gamma source position is not unique, one should be provided with --source-alt")

    return source_alt, source_az


def main():
    log = logging.getLogger("lstchain MC DL2 to IRF - sensitivity curves")

    parser = argparse.ArgumentParser(description="MC DL2 to IRF")

    # Required arguments
    parser.add_argument(
        "--gamma-dl2", "-g", type=str, dest="gamma_file", help="Path to the dl2 gamma file"
    )

    parser.add_argument(
        "--proton-dl2",
        "-p",
        type=str,
        dest="proton_file",
        help="Path to the dl2 proton file",
    )

    parser.add_argument(
        "--electron-dl2",
        "-e",
        type=str,
        dest="electron_file",
        help="Path to the dl2 electron file",
    )

    parser.add_argument(
        "--outfile",
        "-o",
        action="store",
        type=str,
        dest="outfile",
        help="Path where to save IRF FITS file",
        default="sensitivity.fits.gz",
    )

    parser.add_argument(
        "--source_alt",
        action="store",
        type=float,
        dest="source_alt",
        help="Source altitude (optional). If not provided, it will be guessed from the gammas true altitude",
        default=None
    )

    parser.add_argument(
        "--source_az",
        action="store",
        type=float,
        dest="source_az",
        help="Source azimuth (optional). If not provided, it will be guessed from the gammas true altitude",
        default=None
    )

    # Optional arguments
    # parser.add_argument('--config', '-c', action='store', type=Path,
    #                     dest='config_file',
    #                     help='Path to a configuration file. If none is given, a standard configuration is applied',
    #                     default=None
    #                     )


    args = parser.parse_args()
    
    logging.basicConfig(level=logging.INFO)
    logging.getLogger("pyirf").setLevel(logging.DEBUG)

    particles = {
    "gamma": {"file": args.gamma_file, "target_spectrum": CRAB_HEGRA},
    "proton": {"file": args.proton_file, "target_spectrum": IRFDOC_PROTON_SPECTRUM},
    "electron": {
        "file": args.electron_file,
        "target_spectrum": IRFDOC_ELECTRON_SPECTRUM,
    },
}

    for particle_type, p in particles.items():
        log.info("Simulated Events: {}".format(particle_type.title()))
        p["events"], p["simulation_info"] = read_mc_dl2_to_QTable(p["file"])
        # p['events'] = filter_events(p['events'], filters)

        print("=====", particle_type, "=====")
        # p["events"]["particle_type"] = particle_type

        p["simulated_spectrum"] = PowerLaw.from_simulation(p["simulation_info"], T_OBS)
        p["events"]["weight"] = calculate_event_weights(
            p["events"]["true_energy"], p["target_spectrum"], p["simulated_spectrum"]
        )
        for prefix in ("true", "reco"):
            k = f"{prefix}_source_fov_offset"
            p["events"][k] = calculate_source_fov_offset(p["events"], prefix=prefix)

    gammas = particles["gamma"]["events"]
    # background table composed of both electrons and protons
    background = table.vstack(
        [particles["proton"]["events"], particles["electron"]["events"]]
    )

    if args.source_alt is None or args.source_az is None:
        source_alt, source_az = determine_source_position(gammas)
    else:
        source_alt, source_az = args.source_alt, args.source_az

    for particle_type, p in particles.items():
        # calculate theta / distance between reco and assumed source position
        # we handle only ON observations here, so the assumed source pos is the pointing position
        p["events"]["theta"] = calculate_theta(p["events"], assumed_source_az=source_az, assumed_source_alt=source_alt)
        log.info(p["simulation_info"])
        log.info("")


    INITIAL_GH_CUT = np.quantile(gammas["gh_score"], (1 - INITIAL_GH_CUT_EFFICENCY))
    log.info("Using fixed G/H cut of {} to calculate theta cuts".format(INITIAL_GH_CUT))

    # event display uses much finer bins for the theta cut than
    # for the sensitivity
    theta_bins = add_overflow_bins(
        create_bins_per_decade(MIN_ENERGY, MAX_ENERGY, N_BIN_PER_DECADE)
    )

    # theta cut is 68 percent containment of the gammas
    # for now with a fixed global, unoptimized score cut
    mask_theta_cuts = gammas["gh_score"] >= INITIAL_GH_CUT
    theta_cuts = calculate_percentile_cut(
        gammas["theta"][mask_theta_cuts],
        gammas["reco_energy"][mask_theta_cuts],
        bins=theta_bins,
        min_value=MIN_THETA_CUT,
        fill_value=MAX_THETA_CUT,
        max_value=MAX_THETA_CUT,
        percentile=68,
    )

    # same number of bins per decade than official CTA IRFs
    sensitivity_bins = add_overflow_bins(
        create_bins_per_decade(MIN_ENERGY, MAX_ENERGY, bins_per_decade=N_BIN_PER_DECADE)
    )

    log.info("Optimizing G/H separation cut for best sensitivity")
    gh_cut_efficiencies = np.arange(
        GH_CUT_EFFICIENCY_STEP,
        MAX_GH_CUT_EFFICIENCY + GH_CUT_EFFICIENCY_STEP / 2,
        GH_CUT_EFFICIENCY_STEP,
    )
    sensitivity_step_2, gh_cuts = optimize_gh_cut(
        gammas,
        background,
        reco_energy_bins=sensitivity_bins,
        gh_cut_efficiencies=gh_cut_efficiencies,
        op=operator.ge,
        theta_cuts=theta_cuts,
        alpha=ALPHA,
        background_radius=MAX_BG_RADIUS,
    )

    # now that we have the optimized gh cuts, we recalculate the theta
    # cut as 68 percent containment on the events surviving these cuts.
    log.info("Recalculating theta cut for optimized GH Cuts")
    for tab in (gammas, background):
        tab["selected_gh"] = evaluate_binned_cut(
            tab["gh_score"], tab["reco_energy"], gh_cuts, operator.ge
        )

    theta_cuts_opt = calculate_percentile_cut(
        gammas[gammas["selected_gh"]]["theta"],
        gammas[gammas["selected_gh"]]["reco_energy"],
        theta_bins,
        percentile=68,
        fill_value=MAX_THETA_CUT,
        max_value=MAX_THETA_CUT,
        min_value=MIN_THETA_CUT,
    )

    gammas["selected_theta"] = evaluate_binned_cut(
        gammas["theta"], gammas["reco_energy"], theta_cuts_opt, operator.le
    )
    gammas["selected"] = gammas["selected_theta"] & gammas["selected_gh"]

    # calculate sensitivity
    signal_hist = create_histogram_table(
        gammas[gammas["selected"]], bins=sensitivity_bins
    )
    background_hist = estimate_background(
        background[background["selected_gh"]],
        reco_energy_bins=sensitivity_bins,
        theta_cuts=theta_cuts_opt,
        alpha=ALPHA,
        background_radius=MAX_BG_RADIUS,
    )
    sensitivity = calculate_sensitivity(signal_hist, background_hist, alpha=ALPHA)

    # scale relative sensitivity by Crab flux to get the flux sensitivity
    spectrum = particles["gamma"]["target_spectrum"]
    for s in (sensitivity_step_2, sensitivity):
        s["flux_sensitivity"] = s["relative_sensitivity"] * spectrum(
            s["reco_energy_center"]
        )

    log.info("Calculating IRFs")
    hdus = [
        fits.PrimaryHDU(),
        fits.BinTableHDU(sensitivity, name="SENSITIVITY"),
        fits.BinTableHDU(sensitivity_step_2, name="SENSITIVITY_STEP_2"),
        fits.BinTableHDU(theta_cuts, name="THETA_CUTS"),
        fits.BinTableHDU(theta_cuts_opt, name="THETA_CUTS_OPT"),
        fits.BinTableHDU(gh_cuts, name="GH_CUTS"),
    ]

    masks = {
        "": gammas["selected"],
        "_NO_CUTS": slice(None),
        "_ONLY_GH": gammas["selected_gh"],
        "_ONLY_THETA": gammas["selected_theta"],
    }

    # binnings for the irfs
    true_energy_bins = add_overflow_bins(
        create_bins_per_decade(MIN_ENERGY, MAX_ENERGY, N_BIN_PER_DECADE)
    )
    reco_energy_bins = add_overflow_bins(
        create_bins_per_decade(MIN_ENERGY, MAX_ENERGY, N_BIN_PER_DECADE)
    )

    fov_offset_bins = [0, 0.6] * u.deg
    source_offset_bins = np.arange(0, 1 + 1e-4, 1e-3) * u.deg
    energy_migration_bins = np.geomspace(0.2, 5, 200)

    for label, mask in masks.items():
        effective_area = effective_area_per_energy(
            gammas[mask],
            particles["gamma"]["simulation_info"],
            true_energy_bins=true_energy_bins,
        )
        hdus.append(
            create_aeff2d_hdu(
                effective_area[..., np.newaxis],  # add one dimension for FOV offset
                true_energy_bins,
                fov_offset_bins,
                extname="EFFECTIVE_AREA" + label,
            )
        )
        edisp = energy_dispersion(
            gammas[mask],
            true_energy_bins=true_energy_bins,
            fov_offset_bins=fov_offset_bins,
            migration_bins=energy_migration_bins,
        )
        hdus.append(
            create_energy_dispersion_hdu(
                edisp,
                true_energy_bins=true_energy_bins,
                migration_bins=energy_migration_bins,
                fov_offset_bins=fov_offset_bins,
                extname="ENERGY_DISPERSION" + label,
            )
        )

    bias_resolution = energy_bias_resolution(
        gammas[gammas["selected"]],
        true_energy_bins,
        resolution_function=energy_resolution_absolute_68,
    )
    ang_res = angular_resolution(gammas[gammas["selected_gh"]], true_energy_bins)
    psf = psf_table(
        gammas[gammas["selected_gh"]],
        true_energy_bins,
        fov_offset_bins=fov_offset_bins,
        source_offset_bins=source_offset_bins,
    )

    background_rate = background_2d(
        background[background["selected_gh"]],
        reco_energy_bins,
        fov_offset_bins=np.arange(0, 11) * u.deg,
        t_obs=T_OBS,
    )

    hdus.append(
        create_background_2d_hdu(
            background_rate, reco_energy_bins, fov_offset_bins=np.arange(0, 11) * u.deg
        )
    )
    hdus.append(
        create_psf_table_hdu(psf, true_energy_bins, source_offset_bins, fov_offset_bins)
    )
    hdus.append(
        create_rad_max_hdu(
            theta_cuts_opt["cut"][:, np.newaxis], theta_bins, fov_offset_bins
        )
    )
    hdus.append(fits.BinTableHDU(ang_res, name="ANGULAR_RESOLUTION"))
    hdus.append(fits.BinTableHDU(bias_resolution, name="ENERGY_BIAS_RESOLUTION"))

    log.info("Writing output file")
    Path(args.outfile).parent.mkdir(exist_ok=True)
    fits.HDUList(hdus).writeto(args.outfile, overwrite=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
from os import environ
from os.path import join, basename
import subprocess


def main():
    parser = argparse.ArgumentParser(
        description="Batches the r0_to_dl1 lstchain stage for all the files "
        "within a text file."
    )
    parser.add_argument(
        "--file_list",
        "-f",
        type=str,
        dest="file_list",
        help="Path to file containing list of DL0 lstchain file to be processed.",
        required=True,
        nargs="+",
    )
    parser.add_argument(
        "--output_dir",
        "-o",
        type=str,
        dest="output_dir",
        help="lstchain_mc_r0_to_dl1 output directory argument.",
        required=True,
    )
    parser.add_argument(
        "--config_file",
        "-c",
        type=str,
        dest="config_file",
        help="lstchain_mc_r0_to_dl1 configuration file argument.",
        required=True,
    )
    args = parser.parse_args()

    task_id = int(environ["SLURM_ARRAY_TASK_ID"])
    file_for_this_job = args.file_list[task_id]
    print("Using file: ", file_for_this_job)

    # lstchain takes the output dir and constructs filenanmes itself
    with open(file_for_this_job, "r") as filelist:
        for file in filelist:
            file = file.strip("\n")
            # TODO: We need ways to set the pedestal cleaning!
            # or do we? is that not only for observed data?
            output = join(args.output_dir, basename(file))
            # TODO --no-image should be passed by default. To be configurable from lstmcpipe config ?
            cmd = ["lstchain_dl1ab", "--no-image", f"--input-file={file}", f"--output-file={output}"]
            if args.config_file:
                cmd.append("--config={}".format(args.config_file))

            subprocess.run(cmd)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import shutil
from pathlib import Path
from lstmcpipe import prod_logs, __version__


def backup_log(file):
    """
    Backups a log file.
    This scenario should only happens if a MC prod is relaunched because it failed.

    Parameters
    ----------
    file : Path
        Path object to the file to be 'backup-ed'
    """
    if file.exists():
        counter = 0
        save_file = Path(file.parent, f"BACKUP_{counter:02d}_{file.name}")

        while save_file.exists():
            counter += 1
            save_file = Path(file.parent, f"BACKUP_{counter:02d}_{file.name}")

        shutil.copyfile(file, save_file)


def create_log_dir(prod_id):
    """

    Parameters
    ----------
    prod_id : str
        production identifier of the MC production to be launched

    Returns
    -------
    log_dir : Path
        Path to `prod_id` log directory
    """
    log_dir = prod_logs.joinpath(f"logs_{prod_id}")
    log_dir.mkdir(exist_ok=True, parents=True)
    return log_dir


def create_log_files(production_id):
    """
    Manages filenames (and overwrites if needed) log files.

    Parameters
    ----------
    production_id : str
        production identifier of the MC production to be launched

    Returns
    -------
    logs_files: dict
        Dictionary containing
            log_file: Path - path and filename of full log file
            debug_file: Path - path and filename of reduced (debug) log file
    scancel_file: Path
        path and filename of bash file to cancel all the scheduled jobs
    lstmcpipe_log_dir : Path
        Path to prod_ID log directory
    """
    lstmcpipe_log_dir = create_log_dir(production_id)
    log_file = lstmcpipe_log_dir.joinpath(f"log_lstmcpipe_v{__version__}_{production_id}.yml")
    debug_file = lstmcpipe_log_dir.joinpath(f"log_reduced_{production_id}.yml")
    scancel_file = lstmcpipe_log_dir.joinpath(f"scancel_{production_id}.sh")

    # scancel prod file needs chmod +x rights !
    if scancel_file.exists():
        backup_log(scancel_file)
        scancel_file.unlink()
    scancel_file.touch()
    scancel_file.chmod(0o755)  # -rwxr-xr-x

    # If the file exists, i,e., the pipeline has been relaunched, save logs erase it
    if log_file.exists():
        backup_log(log_file)
        log_file.unlink()
    if debug_file.exists():
        backup_log(debug_file)
        debug_file.unlink()

    logs_files = {
        "log_file": log_file,
        "debug_file": debug_file
    }

    return logs_files, scancel_file, lstmcpipe_log_dir


def update_scancel_file(scancel_file, jobids_to_update):
    """
    Bash file containing the slurm command to cancel multiple jobs.
    The file will be updated after every batched stage and will be erased in case the whole MC prod succeed without
    errors.

    Parameters
    ----------
    scancel_file: pathlib.Path
        filename that cancels the whole MC production
    jobids_to_update: str
        job_ids to be included into the the file
    """
    if scancel_file.stat().st_size == 0:
        with open(scancel_file, "r+") as f:
            f.write(f"scancel {jobids_to_update}")

    else:
        with open(scancel_file, "a") as f:
            f.write(f",{jobids_to_update}")



import os
from shutil import which
import pkg_resources
from ruamel.yaml import YAML

from .pipeline_config import load_config
from .dl1ab_tuning import create_dl1ab_tuned_config


__all__ = ["load_config", "create_dl1ab_tuned_config"]


def export_env(outdir="."):
    if which("conda") is not None:
        os.system((f"conda env export > {os.path.join(outdir, 'conda_env.yml')}"))
    elif which("pip") is not None:
        os.system(f"python -m pip freeze > {os.path.join(outdir, 'requirements.txt')}")


def base_config():
    base_config_path = pkg_resources.resource_filename('lstmcpipe', 'base_config_lstmcpipe.yaml')
    return YAML().load(open(base_config_path).read())



"""Helpers to setup logging. Copied and adapted from 
https://github.com/fact-project/aict-tools/blob/ee091fc933110ff008e32df02ee12a97e412e8b3/aict_tools/logging.py
"""

import logging


def setup_logging(logfile=None, verbose=False):
    """
    Setup logging using the logging module
    from the python standard library. A handler for 
    the terminal and optionally a file handler get setup, 
    the logger is using INFO or DEBUG depending on the 
    value of `verbose`.
    Numba logs are filtered if they are not WARNING or above,
    because numba creates thousands of lines in the output.

    Parameters:
    -----------
    logfile: str or path-like
        File to save logs to.
    verbose: boolean
        Whether to enable debug logging

    Returns:
    --------
    logging.Logger
    """

    level = logging.INFO
    if verbose is True:
        level = logging.DEBUG

    log = logging.getLogger()
    log.level = level

    stream_formatter = logging.Formatter(
        fmt="%(levelname)s\n%(message)s\n", datefmt="%H:%M:%S"
    )
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(stream_formatter)
    log.addHandler(stream_handler)

    if logfile is not None:
        file_formatter = logging.Formatter(
            fmt="%(asctime)s|%(levelname)s|%(name)s|%(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S",
        )
        file_handler = logging.FileHandler(logfile)
        file_handler.setFormatter(file_formatter)
        log.addHandler(file_handler)
        log.debug("Added logging handler with log file {}".format(logfile))

    # numba produces a ton of logs when set to debug and the solution of using enviroment variables seems to be inconsistent
    # We dont expect to look through thousands of lines of numba logging anytime soon, so it gets ignored explicitly
    logging.getLogger("numba").setLevel(logging.WARNING)

    return log



import pytest
import tempfile
from ruamel.yaml import YAML

from lstmcpipe.config import paths_config
from lstmcpipe.config import pipeline_config

default_stages = [
    "r0_to_dl1",
    "train_test_split",
    "merge_dl1",
    "train_pipe",
    "dl1_to_dl2",
    "dl2_to_sensitivity",
    "dl2_to_irfs",
]

default_particles = [
    "gamma-diffuse",
    "proton",
    "gamma",
    "electron",
    "proton",
]

default_point_source_offs = [
    "off0.0deg",
    "off0.4deg",
]


@pytest.mark.xfail(raises=NotImplementedError)
def test_path_config_generate_fail():
    pcfg = paths_config.PathConfig('v00')
    pcfg.generate()


def test_path_config_generate():
    class NewPathConfig(paths_config.PathConfig):
        def __init__(self, prod_id):
            super().__init__(prod_id)
            self.stages = ['stage1']

        @property
        def stage1(self):
            return {'input': 'input_path', 'output': None}

    pcfg = NewPathConfig('v00')
    pcfg.generate() == {'stage1': {'input': 'input_path', 'output': None}}


def test_path_config_save():
    pcfg = paths_config.PathConfig('v00')
    pcfg.paths = {'a': 'rick'}
    with tempfile.NamedTemporaryFile() as f:
        pcfg.save_yml(f.name, overwrite=True)
        read_cfg = YAML().load(open(f.name).read())
        pipeline_config.load_config(f.name)
    assert read_cfg['prod_type'] == 'PathConfig'
    assert read_cfg['prod_id'] == 'v00'


def test_PathConfigProd5Trans80():
    from lstmcpipe.config.paths_config import PathConfigProd5Trans80
    prod_id = '20210416_v0.7.3_prod5_trans_80_local_taicut_8_4'
    cfg = PathConfigProd5Trans80(prod_id)
    cfg.generate()
    assert cfg.prod_id == prod_id

    assert "/20200629_prod5_trans_80/" in cfg.base_dir
    assert all(stage in default_stages for stage in cfg.stages)
    assert all(particle in default_particles for particle in cfg.particles)
    assert all(offset in default_point_source_offs for offset in cfg.point_src_offsets)
    assert cfg.training_particles == ["gamma-diffuse", "proton"]
    assert cfg.testing_particles == ["gamma", "electron", "proton", "gamma-diffuse"]
    assert cfg.zenith == "zenith_20deg"
    assert cfg.base_dir == \
           "/fefs/aswg/data/mc/{data_level}/20200629_prod5_trans_80/{particle}/{zenith}/south_pointing/{prod_id}"
    assert cfg.dl1_dir("gamma") == \
           f"/fefs/aswg/data/mc/DL1/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg"
    assert cfg.dl2_dir("gamma") == \
           f"/fefs/aswg/data/mc/DL2/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg"
    assert cfg.dl2_output_file("gamma") == \
           f"/fefs/aswg/data/mc/DL2/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg/dl2_gamma_{prod_id}_test.h5"
    assert cfg.irf_dir("gamma") == \
           f"/fefs/aswg/data/mc/IRF/20200629_prod5_trans_80/zenith_20deg/south_pointing/{prod_id}/gamma"
    assert cfg.merge_output_file("gamma", "train") == \
           f"/fefs/aswg/data/mc/DL1/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg/dl1_gamma_{prod_id}_train.h5"
    assert cfg.merge_output_file("gamma", "test") == \
           f"/fefs/aswg/data/mc/DL1/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg/dl1_gamma_{prod_id}_test.h5"
    assert cfg.models_path() == \
           f"/fefs/aswg/data/models/20200629_prod5_trans_80/zenith_20deg/south_pointing/{prod_id}"
    assert cfg.r0_dir("gamma") == \
           "/fefs/aswg/data/mc/DL0/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/off0.4deg"
    assert cfg.sensitivity_file(offset="0.0deg") == \
           f"/fefs/aswg/data/mc/IRF/20200629_prod5_trans_80/zenith_20deg/south_pointing/{prod_id}/0.0deg/sensitivity_{prod_id}_0.0deg.fits.gz"
    assert cfg.train_dir("gamma") == \
           f"/fefs/aswg/data/mc/DL1/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg/train"
    assert cfg.test_dir("gamma") == \
           f"/fefs/aswg/data/mc/DL1/20200629_prod5_trans_80/gamma/zenith_20deg/south_pointing/{prod_id}/off0.4deg/test"

    for stg in cfg.stages:
        prop = getattr(cfg, stg)
        assert isinstance(prop, list)

        for path in prop:
            assert isinstance(path, dict)

            if stg == "train_pipe":
                assert isinstance(path["input"], dict)
                assert all(item in ["gamma", "proton"] for item in path["input"])
            elif stg == "train_test_split":
                assert isinstance(path["output"], dict)
                assert all(item in ["train", "test"] for item in path["output"])
            elif stg == 'dl1_to_dl2':
                assert 'path_model' in path
            elif stg == "dl2_to_sensitivity":
                assert isinstance(path["input"], dict)
                assert all(item in ["gamma_file", "proton_file", "electron_file"] for item in path["input"])
            elif stg == "dl2_to_irfs":
                assert "options" in path
                assert isinstance(path["input"], dict)
                assert all(item in ["gamma_file", "proton_file", "electron_file"] for item in path["input"])




# #!/usr/bin/env python
import argparse
from pathlib import Path
from datetime import date

from lstmcpipe.config import paths_config



class ParseKwargs(argparse.Action):
    """
    Parse a string formatted as `option1=foo` into a dict `{option1: foo}`
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = value


def list_config_classes():
    """
    List the classes in `lstmcpipe.config.paths_config` that inherit from `lstmcpipe.config.paths_config.PathConfig`

    Returns
    -------
    [object]: list of classes
    """
    all_attrs = []
    for att in list(paths_config.__dict__):
        try:
            cls_base = getattr(paths_config, att).__mro__
            if paths_config.PathConfig in cls_base:
                all_attrs.append(att)
        except:
            pass
    return all_attrs


def main():
    parser = argparse.ArgumentParser(description="Generate a lstmcpipe config.")

    # Required arguments
    parser.add_argument(dest='config_class',
                        type=str,
                        help=f'Config class name to use. '
                             f'List of implemented classes: {list_config_classes()}'
                        )

    parser.add_argument('--prod_id',
                        type=str,
                        help='production ID'
                        )

    parser.add_argument('--output', '-o',
                        type=Path,
                        help='Path to the output file to dump the generated config. '
                             'Optional, if not provided, the file is dumped locally',
                        default=None
                        )

    parser.add_argument('--kwargs',
                        nargs='*',
                        action=ParseKwargs,
                        help="optional kwargs for the requested config class. Use as: --kwargs option1=foo option2=bar"
                        )


    args = parser.parse_args()

    if not hasattr(paths_config, args.config_class):
        raise NotImplementedError(f"Config class {args.config_class} not implemented in lstmcpipe.config.paths_config")

    output = f'lstmcpipe_config_{date.today()}_{args.config_class}.yaml' if args.output is None else args.output
    prod_id = 'prod_00' if args.prod_id is None else args.prod_id

    # we get the class from its name and instantiate it with the required args
    if args.kwargs:
        cfg = getattr(paths_config, args.config_class)(prod_id, **args.kwargs)
    else:
        cfg = getattr(paths_config, args.config_class)(prod_id)
    cfg.generate()
    cfg.save_yml(output)

    print(f"Config saved in {output}")


if __name__ == '__main__':
    main()



from .mc_process_dl1 import r0_to_dl1, reprocess_dl1, batch_process_dl1
from .mc_train_test_splitting import batch_train_test_splitting, train_test_split
from .mc_merge_dl1 import merge_dl1, batch_merge_dl1
from .mc_train import train_pipe, batch_train_pipe, batch_plot_rf_features
from .mc_dl1_to_dl2 import dl1_to_dl2, batch_dl1_to_dl2
from .mc_dl2_to_irfs import dl2_to_irfs, batch_dl2_to_irfs
from .mc_dl2_to_sensitivity import dl2_to_sensitivity, batch_dl2_to_sensitivity

__all__ = [
    "r0_to_dl1",
    "reprocess_dl1",
    "batch_process_dl1",
    "batch_train_test_splitting",
    "train_test_split",
    "merge_dl1",
    "batch_merge_dl1",
    "train_pipe",
    "batch_train_pipe",
    "batch_plot_rf_features",
    "dl1_to_dl2",
    "batch_dl1_to_dl2",
    "dl2_to_irfs",
    "batch_dl2_to_irfs",
    "dl2_to_sensitivity",
    "batch_dl2_to_sensitivity",
]



import os
from pathlib import Path
from .version import __version__  # noqa

__version__
lstmcpipe_root_dir = Path(__file__).absolute().parents[1]
prod_logs = Path(os.getenv('LSTMCPIPE_PROD_LOGS')) if os.getenv('LSTMCPIPE_PROD_LOGS') is not None \
    else Path(os.getenv('HOME')).joinpath('LSTMCPIPE_PROD_LOGS')



#!/usr/bin/env python

import os
import shutil
import logging
from pathlib import Path
from lstmcpipe.workflow_management import save_log_to_file
from lstmcpipe.io.data_management import check_and_make_dir_without_verification


log = logging.getLogger(__name__)


def batch_dl1_to_dl2(
    dict_paths,
    config_file,
    jobid_from_training,
    batch_config,
    logs,
):
    """
    Function to batch the dl1_to_dl2 stage once the lstchain train_pipe batched jobs have finished.

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    config_file : str
        Path to a configuration file. If none is given, a standard configuration is applied
    jobid_from_training : str
        string containing the jobid from the jobs batched in the train_pipe stage, to be passed to the
        dl1_to_dl2 function (as a slurm dependency)
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to
        dl1_dl2 function
    logs: dict
        Dictionary with logs files

    Returns
    -------
    jobid_for_dl2_to_dl3 : str
        string containing the jobids to be passed to the next stage of the workflow (as a slurm dependency)

    """

    log_dl1_to_dl2 = {}
    jobid_for_dl2_to_dl3 = []
    debug_log = {}

    log.info("==== START {} ==== \n".format("batch dl1_to_dl2_workflow"))

    for paths in dict_paths:

        job_logs, jobid = dl1_to_dl2(
            paths["input"],
            paths["output"],
            path_models=paths["path_model"],
            config_file=config_file,
            wait_jobid_train_pipe=jobid_from_training,
            batch_configuration=batch_config,
            slurm_options=paths.get("slurm_options", None),
        )

        log_dl1_to_dl2.update(job_logs)

        # Single particle dl1_dl2 jobid to be appended
        jobid_for_dl2_to_dl3.append(jobid)
        debug_log[jobid] = (
            f"dl1_to_dl2 jobid that depends on : {jobid_from_training} training job"
        )

    jobid_for_dl2_to_dl3 = ",".join(jobid_for_dl2_to_dl3)

    save_log_to_file(log_dl1_to_dl2, logs["log_file"], workflow_step="dl1_to_dl2")
    save_log_to_file(debug_log, logs["debug_file"], workflow_step="dl1_to_dl2")

    log.info("==== END {} ====".format("batch dl1_to_dl2_workflow"))

    return jobid_for_dl2_to_dl3


def dl1_to_dl2(
    input_file,
    output_dir,
    path_models,
    config_file,
    wait_jobid_train_pipe=None,
    batch_configuration='',
    slurm_options=None,
):
    """
    Convert onsite files from dl1 to dl2

    Parameters
    ----------
    input_file : str
        FILE DL1 path
    output_dir : str
        DIR Dl2 path
    path_models : str
        DIR trained models path
    config_file : str
        Path to a configuration file. If none is given, a standard configuration is applied
    wait_jobid_train_pipe : str
        Comma-separated string with the batched jobid from the train stage to indicate the
        dependencies of the current job to be batched
    batch_configuration : dict
        Dictionary containing the (full) source_environment and the slurm_account strings
        to be passed to the sbatch commands
        ! NOTE : train_pipe AND dl1_to_dl2 MUST BE RUN WITH THE SAME ENVIRONMENT
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    log_dl1_to_dl2 : dict
        log dictionary containing {jobid: batch_cmd} information

    jobid_dl1_to_dl2 : str
        batched job_id to be passed to later stages

    """
    source_environment = batch_configuration["source_environment"]
    slurm_account = batch_configuration["slurm_account"]

    log.info("Working on DL1 files in {}".format(Path(input_file).parent.as_posix()))

    check_and_make_dir_without_verification(output_dir)
    log.info("Output dir: {}".format(output_dir))

    log_dl1_to_dl2 = {}

    cmd = f"{source_environment} lstchain_dl1_to_dl2 -f {input_file} -p {path_models}" \
          f" -o {output_dir}"

    if config_file is not None:
        cmd += f" -c {Path(config_file).resolve().as_posix()}"

    jobe = Path(output_dir).joinpath("dl1_dl2-%j.e").resolve().as_posix()
    jobo = Path(output_dir).joinpath("dl1_dl2-%j.o").resolve().as_posix()

    # sbatch --parsable --dependency=afterok:{wait_ids_proton_and_gammas} --wrap="{cmd}"
    batch_cmd = "sbatch --parsable"
    if slurm_options is not None:
        batch_cmd += f" {slurm_options}"
    else:
        batch_cmd += "  -p short --mem=32G"
    if slurm_account != "":
        batch_cmd += f" -A {slurm_account}"
    if wait_jobid_train_pipe is not None:
        batch_cmd += f" --dependency=afterok:{wait_jobid_train_pipe}"
    batch_cmd += f' -J dl1_2 -e {jobe} -o {jobo} --wrap="{cmd}"'

    # Batch the job at La Palma
    jobid_dl1_to_dl2 = os.popen(batch_cmd).read().strip("\n")
    log_dl1_to_dl2.update({jobid_dl1_to_dl2: batch_cmd})

    if config_file is not None:
        shutil.copyfile(
            config_file,
            Path(output_dir).joinpath(Path(config_file).name)
        )

    return log_dl1_to_dl2, jobid_dl1_to_dl2



#!/usr/bin/env python

import argparse
import subprocess
from os.path import join, basename
from os import environ


def main():
    parser = argparse.ArgumentParser(
        description="Batches the ctapipe-stage1 for all the files "
        "within a text file."
    )
    parser.add_argument(
        "--file_list",
        "-f",
        type=str,
        dest="file_list",
        help="Path to file containing list of DL0 simtel files to be processed.",
        required=True,
        nargs="+",
    )
    parser.add_argument(
        "--output_dir",
        "-o",
        type=str,
        dest="output_dir",
        help="ctapipe-stage1 output directory argument.",
        required=True,
    )
    parser.add_argument(
        "--config_file",
        "-c",
        type=str,
        dest="config_file",
        help="ctapipe-stage1 configuration file argument.",
        required=True,
    )
    args = parser.parse_args()

    task_id = int(environ["SLURM_ARRAY_TASK_ID"])
    file_for_this_job = args.file_list[task_id]
    print("Using file: ", file_for_this_job)

    with open(file_for_this_job, "r") as filelist:
        for file in filelist:
            file = file.strip("\n")

            # ctapipe takes the output filename
            # so we need to construct it first
            output = join(
                args.output_dir, basename(file.replace(".simtel.gz", ".dl1.h5"))
            )

            cmd = ["ctapipe-stage1", f"--input={file}", f"--output={output}"]
            if args.config_file:
                cmd.append("--config={}".format(args.config_file))
            subprocess.run(cmd)



# #!/usr/bin/env python
import os
import logging
import argparse
from pathlib import Path

import ctaplot
import matplotlib.pyplot as plt
from lstmcpipe.plots.plot_irfs import plot_summary_from_file


def plot_comparison(filelist, outfile=None, cta_north=False):
    """
    Create a 2x2 plot comparing different sensitivity curves

    Parameters
    ----------
    filelist: list
        File list with sensitivity curves to be compared.
    outfile: str or Path or None
        path to the output file to be saved.
        if None, the figure is not saved
    cta_north: Bool
        Flag to superpose/add (True) or not (False - Default) the CTA North.
        Imported from ctaplot
    """
    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger("lstchain MC DL2 to IRF - sensitivity curves")

    log.info("Starting lstmcpipe compare irfs script")

    fig, axes = plt.subplots(2, 2, figsize=(15, 10))

    for file in filelist:
        log.info(f"Plotting IRFs from file {file}")
        label = os.path.basename(file)
        plot_summary_from_file(file, axes=axes, label=label)

    if cta_north:
        ctaplot.plot_sensitivity_cta_performance('north', ax=axes.ravel()[0])
        ctaplot.plot_angular_resolution_cta_performance('north', ax=axes.ravel()[1])
        ctaplot.plot_energy_resolution_cta_performance('north', ax=axes.ravel()[2])
        ctaplot.plot_effective_area_cta_performance('north', ax=axes.ravel()[3])

    fig.tight_layout()
    if outfile is not None:
        fig.savefig(outfile, dpi=200, bbox_inches='tight')
    else:
        fig.show()

    return axes


def main():
    parser = argparse.ArgumentParser(description="Produce IRFs comparative plots")

    # Required arguments
    parser.add_argument('--filelist', '-f',
                        type=str,
                        nargs='*',
                        dest='filelist',
                        help='List of IRF files',
                        required=True
                        )

    # optional
    parser.add_argument('--outfile', '-o', action='store', type=Path,
                        dest='outfile',
                        help='Path of the outfile',
                        default='compare_irfs.png',
                        )

    parser.add_argument('--add-cta-north',
                        action='store_true',
                        dest='cta_north',
                        help='add CTA north performances curves'
                        )

    args = parser.parse_args()
    plot_comparison(args.filelist, args.outfile,  args.cta_north)


if __name__ == '__main__':
    main()



# library of functions used for LST analysis data management

import os
import sys
import shutil
from distutils.util import strtobool


def query_yes_no(question, default="yes"):
    """
    Ask a yes/no question via raw_input() and return their answer.

    Parameters
    ----------
    question: str
        question to the user
    default: str - "yes", "no" or None
        resumed answer if the user just hits <Enter>.
        "yes" or "no" will set a default answer for the user
        None will require a clear answer from the user
    Returns
    -------
    bool - True for "yes", False for "no"
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        else:
            try:
                return bool(strtobool(choice))
            except:
                sys.stdout.write(
                    "Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n"
                )


def query_continue(question, default="no"):
    """
    Ask a question and if the answer is no, exit the program.
    Calls `query_yes_no`.

    Parameters
    ----------
    question: str
    default: str

    Returns
    -------
    bool - answer from query_yes_no
    """
    answer = query_yes_no(question, default=default)
    if not answer:
        sys.exit("Program stopped by user")
    else:
        return answer


def check_data_path(data_path):
    """
    Check if the path to some data exists.
    Raise an Error if the path does not exist, is not a directory or does not contain data.

    Parameters
    ----------
    data_path: str
    """
    if not os.path.exists(data_path):
        raise ValueError("The input directory must exist")
    if get_input_filelist(data_path) == []:
        raise ValueError("The input directory is empty")
    # if not data_path.startswith(BASE_DIR):
    #     raise ValueError("The root directory for the data is supposed to be {}".format(BASE_DIR))


def get_input_filelist(data_path):
    """
    Return list of files in `data_path`

    Parameters
    ----------
    data_path: str

    Returns
    -------
    list of str
    """
    return [
        os.path.abspath(os.path.join(data_path, f))
        for f in os.listdir(data_path)
        if os.path.isfile(os.path.join(data_path, f))
    ]


def check_and_make_dir(directory):
    """
    Check if a directory exists or contains data before to makedir.
    If exists, query the user to remove its content.

    Parameters
    ----------
    directory: str
        path to a directory
    """
    if os.path.exists(directory) and os.listdir(directory) != []:
        clean = query_continue(
            "The directory {} is not empty. Do you want to remove its content?".format(
                directory
            ),
            default="no",
        )
        if clean:
            shutil.rmtree(directory)
    os.makedirs(directory, exist_ok=True)


def check_and_make_dir_without_verification(directory):
    if os.path.exists(directory) and os.listdir(directory) != []:
        shutil.rmtree(directory)
    os.makedirs(directory, exist_ok=True)


def check_files_in_dir_from_file(directory, file):
    """
    Check that a list of files from a file exist in a dir

    Parameters
    ----------
    directory
    file

    Returns
    -------

    """
    with open(file) as f:
        lines = f.readlines()

    files_in_dir = os.listdir(directory)
    files_not_in_dir = []
    for line in lines:
        filename = os.path.basename(line.rstrip("\n"))
        if filename not in files_in_dir:
            files_not_in_dir.append(filename)

    return files_not_in_dir


def read_lines_file(file):
    with open(file) as f:
        lines = [line.rstrip("\n") for line in f]
    return lines


def move_dir_content(src, dest):
    files = os.listdir(src)
    for f in files:
        shutil.move(os.path.join(src, f), dest)
    os.rmdir(src)


if __name__ == "__main__":
    pass



#!/usr/bin/env python

import os
from ruamel.yaml import YAML
from datetime import date
from copy import deepcopy
import re

from . import base_config
from ..version import __version__


class PathConfig:
    """
    Base class to generate a Path configuration for a production
    """

    def __init__(self, prod_id):
        self.prod_id = prod_id
        self.paths = {}
        self.stages = []

    def generate(self):
        """
        Generate the stages paths. They are copied into self.paths and returned.

        Returns
        -------
        dict: paths config
        """
        for stage in self.stages:
            if not hasattr(self, stage):
                raise NotImplementedError(f"The stage {stage} is not implemented for this class")
            self.paths[stage] = getattr(self, stage)
        return self.paths

    def save_yml(self, filename, overwrite=False, append=False):
        """
        Dump the path config to a file

        Parameters
        ----------
        filename: str or Path
        overwrite: bool
        append: bool
        """
        config_to_save = base_config()
        config_to_save.yaml_set_start_comment(f"lstmcpipe generated config from {self.__class__.__name__} "
                                              f"- {date.today()}\n\n")

        if self.paths == {}:
            raise ValueError("Empty paths, generate first")

        config_to_save['lstmcpipe_version'] = __version__
        config_to_save['prod_type'] = self.__class__.__name__
        config_to_save['prod_id'] = self.prod_id
        config_to_save['stages_to_run'] = self.stages
        config_to_save['stages'] = self.paths

        if os.path.exists(filename) and not (overwrite or append):
            raise FileExistsError(f"{filename} exists. Set overwrite=True or append=True")
        if append and overwrite:
            raise ValueError("Append or overwrite, not both ;-)")
        if append:
            config_to_save.update(YAML().load((open(filename).read())))

        with open(filename, 'w') as f:
            yaml = YAML()
            yaml.dump(config_to_save, f)


class PathConfigProd5Trans80(PathConfig):
    """
    Standard paths configuration for a prod5_trans_80 MC production
    """

    def __init__(self, prod_id, zenith='zenith_20deg'):
        super().__init__(prod_id)
        self.prod_id = prod_id
        self.zenith = zenith
        self.base_dir = \
            '/fefs/aswg/data/mc/{data_level}/20200629_prod5_trans_80/{particle}/{zenith}/south_pointing/{prod_id}'
        self.training_particles = ['gamma-diffuse', 'proton']
        self.testing_particles = ['gamma', 'electron', 'proton', 'gamma-diffuse']
        self.point_src_offsets = ['off0.0deg', 'off0.4deg']
        self.paths = {}
        self.stages = [
            'r0_to_dl1',
            'train_test_split',
            'merge_dl1',
            'train_pipe',
            'dl1_to_dl2',
            'dl2_to_sensitivity',
            'dl2_to_irfs',
        ]

    @property
    def particles(self):
        return list(set(self.training_particles + self.testing_particles))

    def _data_level_dir(self, prod_id, data_level, particle, gamma_src_offset='off0.4deg'):
        """

        Parameters
        ----------
        data_level: str
            `DL0` or `DL1` or `DL2`
        particle: str
            `proton`, `gamma-diffuse`, `gamma` or `electron`
        gamma_src_offset: str
            for point source gammas only. `off0.0deg` or `off0.4deg`
        Returns
        -------
        str: path to directory
        """
        base = self.base_dir.format(
            data_level=data_level,
            particle=particle,
            zenith=self.zenith,
            prod_id=prod_id
        )
        if particle == 'gamma':
            base = os.path.join(base, gamma_src_offset)
        return base

    def r0_dir(self, particle, gamma_src_offset='off0.4deg'):
        # for R0 dir there is no `prod_id` in the path
        return os.path.realpath(self._data_level_dir(data_level='DL0',
                                                     particle=particle,
                                                     gamma_src_offset=gamma_src_offset,
                                                     prod_id='')
                                )

    def dl1_dir(self, particle, gamma_src_offset='off0.4deg'):
        return self._data_level_dir(data_level='DL1',
                                    particle=particle,
                                    gamma_src_offset=gamma_src_offset,
                                    prod_id=self.prod_id)

    def dl2_dir(self, particle, gamma_src_offset='off0.4deg'):
        return self._data_level_dir(data_level='DL2',
                                    particle=particle,
                                    gamma_src_offset=gamma_src_offset,
                                    prod_id=self.prod_id)

    def irf_dir(self, gamma_src_offset='diffuse'):
        return os.path.join(
            os.path.realpath(self._data_level_dir(data_level='IRF',
                                                  particle='',
                                                  gamma_src_offset=gamma_src_offset,
                                                  prod_id=self.prod_id)
                             ),
            gamma_src_offset
        )

    @property
    def r0_to_dl1(self):
        paths = []
        for particle in self.particles:
            if particle == 'gamma':
                for offset in self.point_src_offsets:
                    r0 = self.r0_dir(particle=particle, gamma_src_offset=offset)
                    dl1 = self.dl1_dir(particle=particle, gamma_src_offset=offset)
                    paths.append({'input': r0, 'output': dl1})
            else:
                r0 = self.r0_dir(particle=particle)
                dl1 = self.dl1_dir(particle=particle)
                paths.append({'input': r0, 'output': dl1})
        return paths

    def train_dir(self, particle, gamma_src_offset='off0.4deg'):
        return os.path.join(
            self.dl1_dir(particle, gamma_src_offset=gamma_src_offset),
            'train'
        )

    def test_dir(self, particle, gamma_src_offset='off0.4deg'):
        return os.path.join(
            self.dl1_dir(particle, gamma_src_offset=gamma_src_offset),
            'test'
        )

    @property
    def train_test_split(self):
        paths = []
        for particle in self.particles:
            if particle == 'gamma':
                for offset in self.point_src_offsets:
                    dl1 = self.dl1_dir(particle=particle, gamma_src_offset=offset)
                    paths.append(
                        {'input': dl1,
                         'output': {
                             'train': self.train_dir(particle),
                             'test': self.test_dir(particle)}
                         })
            else:
                dl1 = self.dl1_dir(particle=particle)
                paths.append({
                    'input': dl1,
                    'output': {
                        'train': self.train_dir(particle),
                        'test': self.test_dir(particle)
                    }
                })
        return paths

    def merge_output_file(self, particle, step, gamma_src_offset='off0.4deg'):
        if not step in ['train', 'test']:
            raise ValueError("Only steps accepted: train or test")
        dl1 = self.dl1_dir(particle=particle, gamma_src_offset=gamma_src_offset)
        return os.path.join(dl1, f'dl1_{particle}_{self.prod_id}_{step}.h5')

    @property
    def merge_dl1(self):
        paths = []
        for particle in self.training_particles:
            if particle == 'gamma':
                for offset in self.point_src_offsets:
                    train = self.train_dir(particle)
                    output_file = self.merge_output_file(particle=particle, step='train', gamma_src_offset=offset)
                    paths.append({
                        'input': train,
                        'output': output_file,
                        'options': '--no-image'
                    })
            else:
                train = self.train_dir(particle)
                output_file = self.merge_output_file(particle=particle, step='train')
                paths.append({
                    'input': train,
                    'output': output_file,
                    'options': '--no-image'
                })

        for particle in self.testing_particles:
            if particle == 'gamma':
                for offset in self.point_src_offsets:
                    test = self.test_dir(particle)
                    output_file = self.merge_output_file(particle=particle, step='test', gamma_src_offset=offset)
                    paths.append({
                        'input': test,
                        'output': output_file
                    })
            else:
                test = self.test_dir(particle)
                output_file = self.merge_output_file(particle=particle, step='test')
                paths.append({
                    'input': test,
                    'output': output_file
                })

        self.paths['merge_dl1'] = paths
        return paths

    def models_path(self):
        p = self.base_dir.format(
            data_level='models',
            particle='',
            zenith=self.zenith,
            prod_id=self.prod_id
        ).replace('/mc/', '/')

        return os.path.realpath(p)

    @property
    def train_pipe(self):
        paths = [{
            'input': {
                'gamma': self.merge_output_file('gamma-diffuse', 'train'),
                'proton': self.merge_output_file('proton', 'train')
            },
            'output': self.models_path()
        }]
        return paths

    @property
    def dl1_to_dl2(self):
        paths = []
        for particle in self.testing_particles:
            if particle == 'gamma':
                for offset in self.point_src_offsets:
                    dl1 = self.merge_output_file(particle=particle, step='test', gamma_src_offset=offset)
                    dl2 = self.dl2_dir(particle, gamma_src_offset=offset)
                    paths.append({
                        'input': dl1,
                        'path_model': self.models_path(),
                        'output': dl2
                    })
            else:
                dl1 = self.merge_output_file(particle=particle, step='test')
                dl2 = self.dl2_dir(particle)
                paths.append({
                    'input': dl1,
                    'path_model': self.models_path(),
                    'output': dl2
                })
        return paths

    def dl2_output_file(self, particle, gamma_src_offset='off0.4deg'):
        dl2_filename = os.path.basename(self.merge_output_file(particle=particle,
                                                               step='test',
                                                               gamma_src_offset=gamma_src_offset)).replace('dl1', 'dl2')
        return os.path.join(
            self.dl2_dir(particle=particle, gamma_src_offset=gamma_src_offset),
            dl2_filename
        )

    def sensitivity_file(self, offset):
        return os.path.join(
            self.irf_dir(gamma_src_offset=offset),
            f'sensitivity_{self.prod_id}_{offset}.fits.gz'
        )

    @property
    def dl2_to_sensitivity(self):
        paths = []

        def path_dict(gamma_part, offset):
            d = {
                'input': {
                    'gamma_file': self.dl2_output_file(gamma_part),
                    'proton_file': self.dl2_output_file('proton'),
                    'electron_file': self.dl2_output_file('electron')
                },
                'output': self.sensitivity_file(offset)
            }
            return d

        # sensitivity can be computed only on point source gammas at the moment
        for offset in self.point_src_offsets:
            paths.append(path_dict('gamma', offset))

        return paths

    @property
    def dl2_to_irfs(self):
        paths = []

        def path_dict(gamma_part, offset):
            d = {
                'input': {
                    'gamma_file': self.dl2_output_file(gamma_part),
                    'proton_file': self.dl2_output_file('proton'),
                    'electron_file': self.dl2_output_file('electron')
                },
                'output': os.path.join(self.irf_dir(gamma_src_offset=offset), f'irf_{self.prod_id}_{offset}.fits.gz'),
                'options': '--point-like' if gamma_part == 'gamma' else ''
            }
            return d

        for gamma_part in ['gamma-diffuse', 'gamma']:
            if gamma_part == 'gamma-diffuse':
                paths.append(path_dict(gamma_part, 'diffuse'))
            else:
                for offset in self.point_src_offsets:
                    paths.append(path_dict(gamma_part, offset))

        return paths

    @property
    def plot_irfs(self):
        paths = []

        def path_dict(offset):
            d = {
                'input': self.sensitivity_file(offset),
                'output': self.sensitivity_file(offset).replace('fits.gz', 'png')
            }
            return d

        for gamma_part in ['gamma-diffuse', 'gamma']:
            if gamma_part == 'gamma-diffuse':
                paths.append(path_dict('diffuse'))
            else:
                for offset in self.point_src_offsets:
                    paths.append(path_dict(offset))
        return paths


class PathConfigProd5Trans80DL1ab(PathConfigProd5Trans80):

    def __init__(self, starting_prod_id, new_prod_id, zenith='zenith_20deg'):
        super(PathConfigProd5Trans80DL1ab, self).__init__(prod_id=new_prod_id, zenith=zenith)
        self.starting_prod_id = starting_prod_id
        self.stages.remove('r0_to_dl1')
        self.stages.remove('train_test_split')
        self.stages.remove('merge_dl1')
        self.stages.insert(0, 'dl1ab')

    def starting_dl1(self, particle, step, gamma_src_offset='off0.4deg'):
        former_merged_dl1 = self.merge_output_file(particle=particle, step=step, gamma_src_offset=gamma_src_offset)
        return former_merged_dl1.replace(self.prod_id, self.starting_prod_id)

    @property
    def dl1ab(self):
        paths = []
        for step in ['train', 'test']:
            for particle in self.particles:
                if particle == 'gamma':
                    for offset in self.point_src_offsets:
                        dl1_input = self.starting_dl1(particle=particle, step=step, gamma_src_offset=offset)
                        dl1_output = self.dl1_dir(particle, gamma_src_offset=offset)
                        paths.append({
                            'input': dl1_input,
                            'output': dl1_output
                        })
                else:
                    dl1_input = self.starting_dl1(particle=particle, step=step, gamma_src_offset='')
                    dl1_output = self.dl1_dir(particle, gamma_src_offset='')
                    paths.append({
                        'input': dl1_input,
                        'output': dl1_output
                    })
        return paths


class PathConfigAllSky(PathConfig):
    """
    Standard paths configuration for a prod5_trans_80 MC production
    """

    def __init__(self, prod_id, dec):
        super().__init__(prod_id)
        self.prod_id = prod_id
        self.base_dir = "/fefs/aswg/data/mc/{data_level}/AllSky/{prod_id}/{particle}/" + dec + "/{pointing}/"
        self.training_dir = "/home/georgios.voutsinas/ws/AllSky/TrainingDataset/{particle}/" + dec + "/sim_telarray/{pointing}/output/"
        self.testing_dir = "/home/georgios.voutsinas/ws/AllSky/TestDataset/sim_telarray/{pointing}/output_v1.4/"

        self.training_particles = ['GammaDiffuse', 'Protons']
        self.testing_particles = ['Crab']

        self.paths = {}
        self.stages = ['r0_to_dl1', 'merge_dl1', 'train_pipe', 'dl1_to_dl2', 'dl2_to_irfs']

    def _search_pointings(self, particle):
        pointing_dirs_ = os.listdir(self.r0_dir(particle=particle, pointing='$$$').split('$$$')[0])
        # Check that pointings contain simtel files
        pointing_dirs = []
        for pointing in pointing_dirs_:
            fullpath = self.r0_dir(particle, pointing)
            if [f for f in os.listdir(fullpath) if f.endswith('.simtel.gz')]:
                pointing_dirs.append(pointing)
        return pointing_dirs

    def training_pointings(self, particle):
        if not hasattr(self, '_training_pointings'):
            try:
                self.load_pointings()
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    "The class must be run on the cluster to load available pointing nodes"
                ) from e
        return self._training_pointings[particle]

    def testing_pointings(self, particle):
        if not hasattr(self, '_testing_pointings'):
            try:
                self.load_pointings()
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    "The class must be run on the cluster to load available pointing nodes"
                ) from e
        return self._testing_pointings[particle]

    def load_pointings(self):
        self._training_pointings = self._get_training_pointings()
        self._testing_pointings = self._get_testing_pointings()

    def _extract_pointing(self, text):
        """
        return a tuple ($0, $1) of pointings based on a text pattern `*_theta_{$0}_az_{$1}_`
        """
        return re.search('.*theta\_(.+?)_az\_(.+?)\_', text)

    def _get_training_pointings(self):
        """
        Find pointings that exist for all training particles
        This is overly complicated because pointings directory names are not consistent particle-wise
        see node_theta_16.087_az_108.090_ vs node_corsika_theta_16.087_az_108.090_
        see testing pointings for a simpler implementation if this get solved
        """
        all_pointings = {}
        for particle in self.training_particles:
            all_pointings[particle] = self._search_pointings(particle)
        intersected_pointings = deepcopy(all_pointings)

        for particle, pointings_text in all_pointings.items():
            for pointing_text in pointings_text:
                pointing_tuple = self._extract_pointing(pointing_text)
                for other_particles, other_pointings_text in all_pointings.items():
                    other_pointings_tuples = [self._extract_pointing(pt) for pt in other_pointings_text]
                    if pointing_tuple not in other_pointings_tuples:
                        if pointing_text in intersected_pointings:
                            intersected_pointings[particle].remove(pointing_text)

        return intersected_pointings

    def _get_testing_pointings(self):
        particle = self.testing_particles[0]
        pointings = set(self._search_pointings(particle))
        for particle in self.testing_particles[1:]:
            pointings.intersection_update(self._search_pointings(particle))
        return {particle: pointings for particle in self.testing_particles}

    def _data_level_dir(self, prod_id, data_level, particle, pointing):
        """

        Parameters
        ----------
        data_level: str
            `DL0` or `DL1` or `DL2`
        particle: str
            `proton`, `gamma-diffuse`, `gamma` or `electron`
        prod_id:
        pointing:

        Returns
        -------
        str: path to directory
        """
        if data_level not in ['models', 'DL1', 'DL2', 'IRF']:
            raise ValueError("data_level should be DL1, DL2 or IRF")
        return self.base_dir.format(data_level=data_level, particle=particle, pointing=pointing, prod_id=prod_id)

    def r0_dir(self, particle, pointing):
        # for R0 dir there is no `prod_id` in the path
        if particle in self.training_particles:
            return self.training_dir.format(particle=particle, pointing=pointing)
        elif particle in self.testing_particles:
            return self.testing_dir.format(pointing=pointing)
        else:
            raise ValueError("unknown particle")

    def dl1_dir(self, particle, pointing):
        return self._data_level_dir(data_level='DL1', particle=particle, pointing=pointing, prod_id=self.prod_id)

    def dl2_dir(self, particle, pointing):
        return self._data_level_dir(data_level='DL2', particle=particle, pointing=pointing, prod_id=self.prod_id)

    def irf_dir(self, pointing):
        return os.path.realpath(self._data_level_dir(data_level='IRF',
                                                     particle='',
                                                     pointing=pointing,
                                                     prod_id=self.prod_id)
                                )

    @property
    def r0_to_dl1(self):
        paths = []
        for particle in self.training_particles:
            for pointing in self.training_pointings(particle):
                r0 = self.r0_dir(particle, pointing)
                dl1 = self.dl1_dir(particle, pointing)
                paths.append({'input': r0, 'output': dl1})
        for particle in self.testing_particles:
            for pointing in self.testing_pointings(particle):
                r0 = self.r0_dir(particle, pointing)
                dl1 = self.dl1_dir(particle, pointing)
                paths.append({'input': r0, 'output': dl1})
        return paths

    def training_merged_dl1(self, particle):
        return os.path.join(
            os.path.realpath(self.dl1_dir(particle, '')),
            f'dl1_{particle}_merged.h5'
        )

    def testing_merged_dl1(self, particle, pointing):
        return os.path.join(
            os.path.realpath(self.dl1_dir(particle, '')),
            f'dl1_{particle}_{pointing}_merged.h5'
        )

    @property
    def merge_dl1(self):
        # for the training particles, all the nodes get merged
        paths = []
        for particle in self.training_particles:
            dl1 = self.dl1_dir(particle, '')
            merged_dl1 = self.training_merged_dl1(particle)
            paths.append({
                'input': dl1,
                'output': merged_dl1,
                'options': '--pattern */*.h5 --no-image',
            })

        # for the testing, we merge per node
        for particle in self.testing_particles:
            for pointing in self.testing_pointings(particle):
                dl1 = self.dl1_dir(particle, pointing)
                merged_dl1 = self.testing_merged_dl1(particle, pointing)
                paths.append({
                    'input': dl1,
                    'output': merged_dl1,
                    'options': '--no-image'
                })

        return paths

    def models_path(self):
        p = self.base_dir.format(data_level='models', particle='', pointing='', prod_id=self.prod_id).replace(
            '/mc/', '/')
        return os.path.realpath(p)

    @property
    def train_pipe(self):
        paths = [{
            'input': {
                'gamma': self.training_merged_dl1('GammaDiffuse'),
                'proton': self.training_merged_dl1('Protons'),
            },
            'output': self.models_path(),
            'slurm_options': '-p long --mem=80G'
        }]
        return paths

    @property
    def dl1_to_dl2(self):
        paths = []
        for particle in self.testing_particles:
            for pointing in self.testing_pointings(particle):
                paths.append({
                    'input': self.testing_merged_dl1(particle, pointing),
                    'path_model': self.models_path(),
                    'output': self.dl2_dir(particle, pointing)
                })
        return paths

    def dl2_output_file(self, particle, pointing):
        filename = os.path.basename(self.testing_merged_dl1(particle, pointing).replace('dl1_', 'dl2_'))
        return os.path.join(self.dl2_dir(particle, pointing), filename)

    @property
    def dl2_to_irfs(self):
        paths = []

        for particle in self.testing_particles:
            for pointing in self.testing_pointings(particle):
                paths.append({
                    'input': {'gamma_file': self.dl2_output_file(particle, pointing),
                              'proton_file': None,
                              'electron_file': None,
                              },
                    'output': os.path.join(self.irf_dir(pointing), f'irf_{self.prod_id}_{pointing}.fits.gz'),
                    'options': '--point-like'
                }
                )

        return paths



#!/usr/bin/env python

import os
import logging
from pathlib import Path
from lstmcpipe.workflow_management import save_log_to_file


log = logging.getLogger(__name__)


def batch_dl2_to_sensitivity(
    dict_paths,
    job_ids_from_dl1_dl2,
    batch_config,
    logs,
):
    """
    Batches the dl2_to_sensitivity stage (`stages.script_dl2_to_sensitivity` based in the pyIRF iib) once the
    dl1_to_dl2 stage had finished.

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    job_ids_from_dl1_dl2: str
        Comma-separated string with the job ids from the dl1_to_dl2 stage to be used as a slurm dependency
        to schedule the current stage
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to
        dl2_to_sensitivity function
    logs: dict
        Dictionary with logs files

    Returns
    -------
    jobid_for_check: str
        Comma-separated jobids batched in the current stage
    """
    log.info("==== START {} ====".format("batch mc_dl2_to_sensitivity"))

    log_dl2_to_sensitivity = {}
    jobid_for_check = []
    debug_log = {}

    for paths in dict_paths:
        job_logs, jobid = dl2_to_sensitivity(
            paths["input"],
            paths["output"],
            batch_configuration=batch_config,
            wait_jobs_dl1_dl2=job_ids_from_dl1_dl2,
            slurm_options=paths.get("slurm_options", None),
        )
        jobid_for_check.append(jobid)
        log_dl2_to_sensitivity.update(job_logs)
        debug_log[jobid] = (
            f"Job_ids from the dl2_to_sensitivity stage and the plot_irfs script that "
            f"depends on the dl1_to_dl2 stage job_ids; {job_ids_from_dl1_dl2}"
        )

    jobid_for_check = ",".join(jobid_for_check)

    save_log_to_file(log_dl2_to_sensitivity, logs["log_file"], "dl2_to_sensitivity")
    save_log_to_file(debug_log, logs["debug_file"], "dl2_to_sensitivity")

    log.info("==== END {} ====".format("batch mc_dl2_to_sensitivity"))

    return jobid_for_check


def dl2_to_sensitivity(
    input_paths,
    output,
    batch_configuration,
    wait_jobs_dl1_dl2,
    slurm_options=None,
):
    """
    Function to run the `script_dl2_to_sensitivity` for the gamma (and the different gamma offsets) and gamma-diffuse
    particles.
    Creates the sensitivity *.fits.gz files and the corresponding sensitivity curve plot.

    Parameters
    ----------
    input_paths: dict
    output: str
    batch_configuration : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to the
        sbatch commands
    wait_jobs_dl1_dl2: str
        Comma-separated string with the jobs (dependency) to wait for before launching the cmd
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    log_dl2_to_sensitivity: dict
        Dictionary with job_id-slurm command key-value pair used for logging
    job_id: str
        String with job_ids batched by the dl2_to_sensitivity script

    """
    log_dl2_to_sensitivity = {}
    jobids_dl2_to_sensitivity = []

    g_file = input_paths["gamma_file"]
    p_file = input_paths["proton_file"]
    e_file = input_paths["electron_file"]

    source_env = batch_configuration["source_environment"]
    slurm_account = batch_configuration["slurm_account"]

    # TODO Move the base commands into scripts so that we can use subprocess properly(makes splitting the string easier)
    base_cmd_sens = f"lstmcpipe_dl2_to_sensitivity -g {g_file} -p {p_file} -e {e_file} -o {output}"

    jobo_sens = Path(output).parent.joinpath("job_dl2_to_sensitivity.o")
    jobe_sens = Path(output).parent.joinpath("job_dl2_to_sensitivity.e")

    cmd_sens = "sbatch --parsable -p short --mem 32G"
    if slurm_account != "":
        cmd_sens += f" -A {slurm_account}"
    cmd_sens += (
        f" --dependency=afterok:{wait_jobs_dl1_dl2} -e {jobe_sens} -o {jobo_sens}"
        f' -J dl2_sens --wrap="{source_env} {base_cmd_sens}"'
    )

    job_id_dl2_sens = os.popen(cmd_sens).read().strip("\n")
    log_dl2_to_sensitivity.update({job_id_dl2_sens: cmd_sens})
    jobids_dl2_to_sensitivity.append(job_id_dl2_sens)

    # Create plot from sensitivity files
    base_cmd_plot = (
        f'lstmcpipe_plot_irfs -f {output} -o {output.replace(".fits.gz", ".png")}'
    )
    jobe_plot = Path(output).parent.joinpath("job_plot_sensitivity-%j.e")
    jobo_plot = Path(output).parent.joinpath("job_plot_sensitivity-%j.o")

    cmd_plot = "sbatch --parsable"
    if slurm_options is not None:
        cmd_plot += f" {slurm_options}"
    else:
        cmd_plot += " -p short"
    if slurm_account != "":
        cmd_plot += f" -A {slurm_account}"
    cmd_plot += (
        f" --dependency=afterok:{job_id_dl2_sens} -e {jobe_plot} -o {jobo_plot}"
        f' -J dl2_sens_plot --wrap="export MPLBACKEND=Agg; {source_env} {base_cmd_plot}"'
    )

    job_id_plot_sens = os.popen(cmd_plot).read().strip("\n")
    log_dl2_to_sensitivity.update({job_id_plot_sens: cmd_plot})
    jobids_dl2_to_sensitivity.append(job_id_plot_sens)

    return log_dl2_to_sensitivity, ",".join(jobids_dl2_to_sensitivity)



#!/usr/bin/env python

# DL2 to IRFs stage for the R0 to IRFs full MC production workflow.
# As all the previous stages, this script make use of the lstchain entry points to batch each workflow stage.

import os
import shutil
import logging
from pathlib import Path
from lstmcpipe.workflow_management import save_log_to_file
from lstmcpipe.io.data_management import check_and_make_dir_without_verification


log = logging.getLogger(__name__)


def batch_dl2_to_irfs(
    dict_paths,
    config_file,
    job_ids_from_dl1_dl2,
    batch_config,
    logs,
):
    """
    Batches the dl2_to_irfs stage (lstchain lstchain_create_irf_files script) once the dl1_to_dl2 stage had finished.

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    config_file: str
        Path to lstchain-like config file
    job_ids_from_dl1_dl2: str
        Comma-separated string with the job ids from the dl1_to_dl2 stage to be used as a slurm dependency
        to schedule the current stage
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to
        dl2_to_irfs function
    logs: dict
        Dictionary with logs files

    Returns
    -------
    jobs_from_dl2_irf: str
        Comma-separated jobids batched in the current stage
    """
    log.info("==== START {} ====".format("batch mc_dl2_to_irfs"))

    log_dl2_to_irfs = {}
    jobid_for_check = []
    debug_log = {}

    for paths in dict_paths:

        job_logs, jobid = dl2_to_irfs(
            paths["input"]["gamma_file"],
            paths["input"]["electron_file"],
            paths["input"]["proton_file"],
            paths["output"],
            config_file=config_file,
            irf_point_like=paths["options"],
            batch_configuration=batch_config,
            wait_jobs_dl1dl2=job_ids_from_dl1_dl2,
            slurm_options=paths.get("slurm_options", None),
        )

        log_dl2_to_irfs.update(log_dl2_to_irfs)
        jobid_for_check.append(jobid)
        debug_log[jobid] = (
            f"jobid from dl2_to_irfs stage that depends of the dl1_to_dl2 stage "
            f"job_ids; {job_ids_from_dl1_dl2}"
        )

    jobid_for_check = ",".join(jobid_for_check)

    save_log_to_file(log_dl2_to_irfs, logs["log_file"], workflow_step="dl2_to_irfs")
    save_log_to_file(debug_log, logs["debug_file"], workflow_step="dl2_to_irfs")

    log.info("==== END {} ====".format("batch mc_dl2_to_irfs"))

    return jobid_for_check


def dl2_to_irfs(
    gamma_file,
    electron_file,
    proton_file,
    outfile,
    config_file,
    irf_point_like,
    batch_configuration,
    wait_jobs_dl1dl2,
    slurm_options=None
):
    """
    Batches interactively the lstchain `lstchain_create_irf_files` entry point.

    Parameters
    ----------
    gamma_file: str
    electron_file: str
    proton_file: str
    outfile: str
    config_file: str
        Path to a configuration file. If none is given, a standard configuration is applied
    irf_point_like: str
        MC prod configuration argument to create IRFs: {True: gamma, False: gamma-diffuse}.
    batch_configuration : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed to the
        sbatch commands
    wait_jobs_dl1dl2: str
        Comma separated string with the job ids of previous stages (dl1_to_dl2 stage) to be passed as dependencies to
        the create_irfs_files job to be batched.
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    log_dl2_to_irfs: dict
        Dictionary-wise log containing {'job_id': 'batched_cmd'} items
    job_id_dl2_irfs: str
        Job-id of the batched job to be passed to the last (MC prod check) stage of the workflow.
    """
    source_env = batch_configuration["source_environment"]
    slurm_account = batch_configuration["slurm_account"]
    
    output_dir = Path(outfile).parent

    log_dl2_to_irfs = {}

    check_and_make_dir_without_verification(output_dir)

    cmd = (
        f"lstchain_create_irf_files {irf_point_like} -g {gamma_file} -o {outfile} "
    )
    if proton_file is not None:
        cmd += f" -p {proton_file}"
    if electron_file is not None:
        cmd += f" -e {electron_file}"

    if config_file:
        cmd += f" --config={config_file}"

    log.info(f"Output dir IRF of {gamma_file}: {output_dir}")

    jobe = Path(output_dir).joinpath("job_dl2_to_irfs-%j.e").resolve().as_posix()
    jobo = Path(output_dir).joinpath("job_dl2_to_irfs-%j.o").resolve().as_posix()

    batch_cmd = "sbatch --parsable"
    if slurm_options is not None:
        batch_cmd += f" {slurm_options}"
    else:
        batch_cmd += " -p short"
    if slurm_account != "":
        batch_cmd += f" -A {slurm_account}"
    batch_cmd += (
        f" --dependency=afterok:{wait_jobs_dl1dl2} -J dl2_IRF"
        f' -e {jobe} -o {jobo} --wrap="{source_env} {cmd}"'
    )

    job_id_dl2_irfs = os.popen(batch_cmd).read().strip("\n")
    log_dl2_to_irfs.update({job_id_dl2_irfs: batch_cmd})

    # Copy config into working dir
    if config_file:
        shutil.copyfile(
            config_file,
            Path(output_dir).joinpath(Path(config_file).name)
        )

    return log_dl2_to_irfs, job_id_dl2_irfs



from lstmcpipe.config import export_env
from lstmcpipe.config.pipeline_config import (
    config_valid,
    complete_lstmcpipe_config,
)
import tempfile
import os
import pytest
# from datetime import datetime

yaml_keys = [
    "prod_id",
    "stages_to_run",
    "prod_type",
    "workflow_kind",
    "stages"
]
dummy_config = {k: None for k in yaml_keys}
dummy_config["merging_options"] = {"no_image": False}
dummy_config["source_environment"] = {"source_file": "src_file", "conda_env": "env"}


def test_export_env():
    tmp_dir = tempfile.mkdtemp()
    export_env(tmp_dir)
    if not os.path.exists(os.path.join(tmp_dir, "requirements.txt")):
        assert os.path.exists(os.path.join(tmp_dir, "conda_env.yml"))


def test_config_valid():
    """
    Test invalid values for all relevant fields individually.
    This tests values, that are supposed to fail.
    This does not test if no valid configuration is marked
    as invalid!
    """
    with pytest.raises(Exception):
        config_valid(dummy_config)

    valid = dummy_config.copy()
    valid["prod_type"] = "PathConfigProd5Trans80"
    valid["workflow_kind"] = "lstchain"
    valid["stages_to_run"] = []
    assert config_valid(valid)

    invalid_workflow = valid.copy()
    invalid_workflow["workflow_kind"] = "MARS"
    with pytest.raises(Exception):
        config_valid(invalid_workflow)

    missing_reference = valid.copy()
    missing_reference["stages_to_run"] = ["dl1ab"]
    with pytest.raises(KeyError):
        config_valid(missing_reference)


def test_complete_lstmcpipe_config():
    """
    Test that the file paths are properly set.
    This only tests a subset of the variables, that are set in
    the function.
    """
    config = dummy_config.copy()
    config["workflow_kind"] = "lstchain"
    config["prod_type"] = "PathConfigProd5Trans80"
    config["stages_to_run"] = ["r0_to_dl1"]

    parsed_config = complete_lstmcpipe_config(config)
    # date = datetime.today().strftime("%Y%m%d")
    # TODO test check some paths
    # TODO test check model path
    assert (
        parsed_config["batch_config"]["source_environment"] == "source src_file; conda activate env; "
    )
    assert parsed_config["batch_config"]["slurm_account"] == ""
    assert parsed_config["stages_to_run"] == ["r0_to_dl1"]
    assert parsed_config["workflow_kind"] == "lstchain"

    new_dummy = config.copy()
    new_dummy["slurm_config"] = {"user_account": "aswg"}
    new_parsed_config = complete_lstmcpipe_config(new_dummy)
    assert new_parsed_config["batch_config"]["slurm_account"] == "aswg"



# this is adapted from https://github.com/astropy/astropy/blob/master/astropy/version.py
# see https://github.com/astropy/astropy/pull/10774 for a discussion on why this needed.

try:
    try:
        from ._dev_version import version
    except ImportError:
        from ._version import __version__ as version
except Exception as e:
    import warnings

    warnings.warn(
        "Could not determine lstmcpipe version; this indicates a broken installation."
        " Install lstmcpipe from PyPI, a local git repository or using "
        " pip install git+https://github.com/cta-observatory/lstmcpipe@VERSION "
        " Installing github's autogenerated source release tarballs "
        " does not include version information and should be avoided."
        f"Exception Error: {e}"
    )
    del warnings
    version = "0.0.0"

__version__ = version


# #!/usr/bin/env python

# E. Garcia, T. VUILLAUME, LAPP, CNRS
#
# Full workflow from r0 to dl3.
#  wraps together the individual scripts of
#   - mc_process_dl1
#   - mc_train_test_split
#   - mc_merge_dl1
#   - mc_train
#   - mc_dl1_to_dl2
#   - mc_dl2_to_irfs
#   - mc_dl2_to_sensitivity
#
# usage:
# > lstmcpipe -c config_MC_prod.yml -conf_lst LSTCHAIN_CONFIG_FILE [-conf_rta RTA_CONFIG_FILE]
#   [-conf_cta CTA_CONFIG_FILE]
#

import argparse
from pathlib import Path
from lstmcpipe.logging import setup_logging
from lstmcpipe.io.data_management import query_continue
from lstmcpipe.config import load_config, create_dl1ab_tuned_config
from lstmcpipe.io.lstmcpipe_tree_path import (
    create_log_files,
    update_scancel_file,
)
from lstmcpipe.workflow_management import (
    batch_mc_production_check,
)
from lstmcpipe.stages import (
    batch_process_dl1,
    batch_train_test_splitting,
    batch_merge_dl1,
    batch_train_pipe,
    batch_dl1_to_dl2,
    batch_dl2_to_irfs,
    batch_dl2_to_sensitivity,
    batch_plot_rf_features,
)


parser = argparse.ArgumentParser(description="MC R0 to DL3 full pipeline")

parser.add_argument(
    "--config_mc_prod",
    "-c",
    action="store",
    type=str,
    dest="config_mc_prod",
    help="Path to the MC_production configuration file. ",
    default="./config_MC_prod.yml",
    required=True,
)

parser.add_argument(
    "--config_file_lst",
    "-conf_lst",
    action="store",
    type=str,
    dest="config_file_lst",
    help="Path to a lstchain-like configuration file. "
    "RF classifier and regressor arguments must be declared here !",
    required=True
)

parser.add_argument(
    "--config_file_ctapipe",
    "-conf_cta",
    action="store",
    type=str,
    dest="config_file_ctapipe",
    help="Path to a ctapipe-like configuration file."
    'Only to be declared if WORKFLOW_KIND = "ctapipe" and only used up to dl1.',
    default=None,
)

parser.add_argument(
    "--config_file_rta",
    "-conf_rta",
    action="store",
    type=str,
    dest="config_file_rta",
    help="Path to a HiPeRTA-like configuration file."
    'Only to be declared if WORKFLOW_KIND = "hiperta" and only used up to dl1.',
    default=None,
)

parser.add_argument(
    "--debug", action="store_true", help="print debug messages to stderr"
)
parser.add_argument(
    "--log-file",
    action="store",
    type=str,
    dest="log_file",
    help="Optional log file. This is independent of the slurm job logs and only handles lstmcpipe logging",
    default=None,
)

args = parser.parse_args()

#######################################################################################################################


def main():
    """
    Main lstmcpipe script. This will launch the selected stages and start the processing.
    The jobs are submitted, but not awaited meaning that the analysis will be going on after
    the script has exited. To look at the submitted jobs, you can use e.g. `squeue -u $USER`.
    
    Arguments, that can be passed via the command line:
    ---------------------------------------------------
    --config_mc_prod / -c
        path to a yaml configuration file containing lstmcpipe settings
        This defines the stages run, the files processed, ...
    --config_file_lst / -conf_lst
        path to a yaml configuration file containing lstchain settings
        This defines the processing parameters like cleaning, models...
    --config_file_ctapipe / -conf_cta
        same for ctapipe
    --config_file_rat / -conf_rta
        same for HIPERTA
    --log-file
        Optional: path to a file where lstmcpipe logging will be written to.
    --debug
        Toggle to enable debug print messages.
    """
    log = setup_logging(verbose=args.debug, logfile=args.log_file)
    log.info("Starting lstmcpipe processing script")
    # Read MC production configuration file
    lstmcpipe_config = load_config(args.config_mc_prod)
    query_continue("Are you sure ?")

    # Load variables
    prod_id = lstmcpipe_config["prod_id"]
    workflow_kind = lstmcpipe_config["workflow_kind"]
    batch_config = lstmcpipe_config["batch_config"]
    stages_to_run = lstmcpipe_config["stages_to_run"]

    # Create log files and log directory
    logs_files, scancel_file, logs_dir = create_log_files(prod_id)
    all_job_ids = {}

    if "r0_to_dl1" in stages_to_run and "reprocess_dl1" in stages_to_run:
        raise Exception("There can only be one stage producing dl1 files")

    # 1 STAGE --> R0/1 to DL1 or reprocessing of existing dl1a files
    r0_to_dl1 = "r0_to_dl1" in stages_to_run
    dl1ab = "dl1ab" in stages_to_run

    if r0_to_dl1 or dl1ab:

        if workflow_kind == "lstchain":
            dl1_config = Path(args.config_file_lst).resolve().as_posix()
            if lstmcpipe_config.get("dl1_noise_tune_data_run"):
                dl1_config = create_dl1ab_tuned_config(
                    dl1_config,
                    dl1_config.replace(".json", "_tuning.json"),
                    lstmcpipe_config.get("dl1_noise_tune_data_run"),
                    lstmcpipe_config.get("dl1_noise_tune_mc_run"),
                )
        elif workflow_kind == "hiperta":
            dl1_config = Path(args.config_file_rta).resolve().as_posix()
        else:  # if this was not ctapipe, the config parsing would have failed
            dl1_config = Path(args.config_file_ctapipe).resolve().as_posix()

        jobs_from_dl1_processing = batch_process_dl1(
            lstmcpipe_config['stages'],
            conf_file=dl1_config,
            batch_config=batch_config,
            workflow_kind=workflow_kind,
            new_production=r0_to_dl1,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_dl1_processing)
        if r0_to_dl1:
            all_job_ids.update({"r0_dl1": jobs_from_dl1_processing})
        else:
            all_job_ids.update({"dl1ab": jobs_from_dl1_processing})
    else:
        jobs_from_dl1_processing = ""

    # 2.1 STAGE --> Train, test splitting
    if "train_test_split" in stages_to_run:
        jobs_from_splitting = batch_train_test_splitting(
            lstmcpipe_config['stages']['train_test_split'],
            jobids_from_r0dl1=jobs_from_dl1_processing,
            batch_config=batch_config,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_splitting)
        all_job_ids.update({"train_test_split": jobs_from_splitting})
    else:
        jobs_from_splitting = ""

    # 2.2 STAGE --> Merge DL1 files
    if jobs_from_splitting != "":
        merge_wait_jobs = ','.join([jobs_from_dl1_processing, jobs_from_splitting])
    else:
        merge_wait_jobs = jobs_from_dl1_processing

    if "merge_dl1" in stages_to_run:
        jobs_from_merge = batch_merge_dl1(
            lstmcpipe_config['stages']["merge_dl1"],
            jobid_from_splitting=merge_wait_jobs,
            batch_config=batch_config,
            workflow_kind=workflow_kind,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_merge)
        all_job_ids.update({"merge_and_copy_dl1": jobs_from_merge})
    else:
        jobs_from_merge = ""

    # 3 STAGE --> Train pipe
    if "train_pipe" in stages_to_run:

        job_from_train_pipe = batch_train_pipe(
            lstmcpipe_config['stages']['train_pipe'],
            jobs_from_merge,
            config_file=Path(args.config_file_lst).resolve().as_posix(),
            batch_config=batch_config,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, job_from_train_pipe)
        all_job_ids.update({"train_pipe": job_from_train_pipe})

        # Plot the RF feature's importance
        job_from_plot_rf_feat = batch_plot_rf_features(
            lstmcpipe_config['stages']['train_pipe'],
            Path(args.config_file_lst).resolve().as_posix(),
            batch_config,
            job_from_train_pipe,
            logs=logs_files,
        )
        update_scancel_file(scancel_file, job_from_plot_rf_feat)
        all_job_ids.update({"plot_rf_feat": job_from_plot_rf_feat})

    else:
        job_from_train_pipe = ""

    # 4 STAGE --> DL1 to DL2 stage
    if "dl1_to_dl2" in stages_to_run:

        jobs_from_dl1_dl2 = batch_dl1_to_dl2(
            lstmcpipe_config['stages']['dl1_to_dl2'],
            Path(args.config_file_lst).resolve().as_posix(),
            job_from_train_pipe,  # Single jobid from train
            batch_config=batch_config,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_dl1_dl2)
        all_job_ids.update({"dl1_to_dl2": jobs_from_dl1_dl2})
    else:
        jobs_from_dl1_dl2 = ""

    # 5 STAGE --> DL2 to IRFs stage
    if "dl2_to_irfs" in stages_to_run:
        jobs_from_dl2_irf = batch_dl2_to_irfs(
            lstmcpipe_config['stages']['dl2_to_irfs'],
            Path(args.config_file_lst).resolve().as_posix(),
            jobs_from_dl1_dl2,
            batch_config=batch_config,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_dl2_irf)
        all_job_ids.update({"dl2_to_irfs": jobs_from_dl2_irf})

    # 6 STAGE --> DL2 to sensitivity curves
    if "dl2_to_sensitivity" in stages_to_run:
        jobs_from_dl2_sensitivity = batch_dl2_to_sensitivity(
            lstmcpipe_config['stages']['dl2_to_sensitivity'],
            jobs_from_dl1_dl2,
            batch_config=batch_config,
            logs=logs_files,
        )

        update_scancel_file(scancel_file, jobs_from_dl2_sensitivity)
        all_job_ids.update({"dl2_to_sensitivity": jobs_from_dl2_sensitivity})

    # Check DL2 jobs and the full workflow if it has finished correctly
    jobid_check = batch_mc_production_check(
        all_job_ids,
        log_directory=logs_dir,
        prod_id=prod_id,
        prod_config_file=args.config_mc_prod,
        batch_config=batch_config,
        logs_files=logs_files,
    )

    update_scancel_file(scancel_file, jobid_check)
    log.info("Finished lstmcpipe processing script. All jobs have been submitted")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Reorganize the DL1 output files of the HiPeRTA and HiPeCTA r0_to_dl1 codes to reach the lstchain
# file structure for DL1 files
#
# Usage :
# $ python reorganize_dl1_file -i input.h5 [-o outname.h5]

import os
import copy
import tables
import argparse
import numpy as np
import pandas as pd
import astropy.units as u
from astropy.table import join, Table, vstack, Column
from ctapipe.coordinates import CameraFrame
from astropy.time import Time
from astropy.coordinates import SkyCoord, EarthLocation, AltAz

# from lstchain.io.io import dl1_params_lstcam_key, dl1_images_lstcam_key, add_column_table
# from lstchain.reco import disp
# from lstchain.reco.utils import sky_to_camera

# Make the hiperta side of lst_scripts independent of lstchian. Thus the reorganizer contains all the functions
#  from lstchain that needs to run

dl1_params_lstcam_key = "dl1/event/telescope/parameters/LST_LSTCam"
dl1_images_lstcam_key = "dl1/event/telescope/image/LST_LSTCam"
# position of the LST1
location = EarthLocation.from_geodetic(-17.89139 * u.deg, 28.76139 * u.deg, 2184 * u.m)
obstime = Time("2018-11-01T02:00")
horizon_frame = AltAz(location=location, obstime=obstime)


parser = argparse.ArgumentParser(
    description="Re-organize the dl1 `standard` output file from either the "
    "hiptecta_r1_to_dl1 or hiperta_r1_dl1 to the lstchain DL1 structure"
)

parser.add_argument(
    "--infile",
    "-i",
    type=str,
    dest="infile",
    help="dl1 output file of the `hipecta_hdf5_r1_to_dl1.py` or `hiperta_r1_dl1` script "
    "to be re-organized",
    default=None,
)

parser.add_argument(
    "--outfile",
    "-o",
    type=str,
    dest="outfile",
    help="Output filename. dl1_reorganized.h5 by default.",
    default="./dl1_reorganized.h5",
)


def add_column_table(table, ColClass, col_label, values):
    """
    FUNCTION COPIED FROM lstchain.io.io to avoid hiperta depend on lstchain.

    Add a column to an pytable Table

    Parameters
    ----------
    table: `tables.table.Table`
    ColClass: `tables.atom.MetaAtom`
    col_label: str
    values: list or `numpy.ndarray`

    Returns
    -------
    `tables.table.Table`
    """
    # Step 1: Adjust table description
    d = table.description._v_colobjects.copy()  # original description
    d[col_label] = ColClass()  # add column

    # Step 2: Create new temporary table:
    newtable = tables.Table(
        table._v_file.root, "_temp_table", d, filters=table.filters
    )  # new table
    table.attrs._f_copy(newtable)  # copy attributes
    # Copy table rows, also add new column values:
    for row, value in zip(table, values):
        newtable.append([tuple(list(row[:]) + [value])])
    newtable.flush()

    # Step 3: Move temporary table to original location:
    parent = table._v_parent  # original table location
    name = table._v_name  # original table name
    table.remove()  # remove original table
    newtable.move(parent, name)  # move temporary table to original location

    return newtable


def disp(cog_x, cog_y, src_x, src_y):
    """
    FUNCTION COPIED FROM lstchain.reco.disp to avoid hiperta depend on lstchain.

    Compute the disp parameters

    Parameters
    ----------
    cog_x: `numpy.ndarray` or float
    cog_y: `numpy.ndarray` or float
    src_x: `numpy.ndarray` or float
    src_y: `numpy.ndarray` or float

    Returns
    -------
    (disp_dx, disp_dy, disp_norm, disp_angle, disp_sign):
        disp_dx: 'astropy.units.m`
        disp_dy: 'astropy.units.m`
        disp_norm: 'astropy.units.m`
        disp_angle: 'astropy.units.rad`
        disp_sign: `numpy.ndarray`
    """
    disp_dx = src_x - cog_x
    disp_dy = src_y - cog_y
    disp_norm = np.sqrt(disp_dx ** 2 + disp_dy ** 2)
    if hasattr(disp_dx, "__len__"):
        disp_angle = np.arctan(disp_dy / disp_dx)
        disp_angle[disp_dx == 0] = np.pi / 2.0 * np.sign(disp_dy[disp_dx == 0])
    else:
        if disp_dx == 0:
            disp_angle = np.pi / 2.0 * np.sign(disp_dy)
        else:
            disp_angle = np.arctan(disp_dy / disp_dx)

    disp_sign = np.sign(disp_dx)

    return disp_dx, disp_dy, disp_norm, disp_angle, disp_sign


def clip_alt(alt):
    """
    FUNCTION COPIED FROM lstchain.reco.utils to avoid hiperta depend on lstchain.

    Make sure altitude is not larger than 90 deg (it happens in some MC files for zenith=0),
    to keep astropy happy
    """
    return np.clip(alt, -90.0 * u.deg, 90.0 * u.deg)


def sky_to_camera(alt, az, focal, pointing_alt, pointing_az):
    """
    FUNCTION COPIED FROM lstchain.reco.utils to avoid hiperta depend on lstchain.

    Coordinate transform from aky position (alt, az) (in angles) to camera coordinates (x, y) in distance
    Parameters
    ----------
    alt: astropy Quantity
    az: astropy Quantity
    focal: astropy Quantity
    pointing_alt: pointing altitude in angle unit
    pointing_az: pointing altitude in angle unit

    Returns
    -------
    camera frame: `astropy.coordinates.sky_coordinate.SkyCoord`
    """
    pointing_direction = SkyCoord(
        alt=clip_alt(pointing_alt), az=pointing_az, frame=horizon_frame
    )

    camera_frame = CameraFrame(
        focal_length=focal, telescope_pointing=pointing_direction
    )

    event_direction = SkyCoord(alt=clip_alt(alt), az=az, frame=horizon_frame)

    camera_pos = event_direction.transform_to(camera_frame)

    return camera_pos


def add_disp_and_mc_type_to_parameters_table(dl1_file, table_path):
    """
    HARDCODED function obtained from `lstchain.reco.dl0_to_dl1` because `mc_alt_tel` and `mc_az_tel` are zipped within
    `run_array_direction`.
    1. Reconstruct the disp parameters and source position from a DL1 parameters table and write the result in the file.
    2. Computes mc_type from the name of the file.


    Parameters
    ----------
    dl1_file: HDF5 DL1 file containing the required field in `table_path`:
        - mc_alt
        - mc_az
        - mc_alt_tel
        - mc_az_tel

    table_path: path to the parameters table in the file

    Returns
    -------
        None
    """
    with tables.open_file(dl1_file) as hfile:
        run_array_dir = copy.copy(
            hfile.root.simulation.run_config.col("run_array_direction")[0]
        )
        # Remember that /telescope has been moved previously
        focal = copy.copy(
            hfile.root.instrument.telescope.optics.col("equivalent_focal_length")[0]
        )

    df = pd.read_hdf(dl1_file, key=table_path)
    source_pos_in_camera = sky_to_camera(
        df.mc_alt.values * u.rad,
        df.mc_az.values * u.rad,
        focal * u.m,
        run_array_dir[1] * u.rad,
        run_array_dir[0] * u.rad,
    )

    disp_parameters = disp(
        df.x.values * u.m,
        df.y.values * u.m,
        source_pos_in_camera.x,
        source_pos_in_camera.y,
    )

    with tables.open_file(dl1_file, mode="a") as file:
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "disp_dx", disp_parameters[0].value)
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "disp_dy", disp_parameters[1].value)
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "disp_norm", disp_parameters[2].value)
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "disp_angle", disp_parameters[3].value)
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "disp_sign", disp_parameters[4])
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "src_x", source_pos_in_camera.x.value)
        tab = file.root[table_path]
        add_column_table(tab, tables.Float32Col, "src_y", source_pos_in_camera.y.value)
        tab = file.root[table_path]
        add_column_table(
            tab, tables.Float32Col, "mc_alt_tel", np.ones(len(df)) * run_array_dir[1]
        )
        tab = file.root[table_path]
        add_column_table(
            tab, tables.Float32Col, "mc_az_tel", np.ones(len(df)) * run_array_dir[0]
        )
        if "gamma" in dl1_file:
            tab = file.root[table_path]
            add_column_table(tab, tables.Float32Col, "mc_type", np.zeros(len(df)))
        if "electron" in dl1_file:
            tab = file.root[table_path]
            add_column_table(tab, tables.Float32Col, "mc_type", np.ones(len(df)))
        if "proton" in dl1_file:
            tab = file.root[table_path]
            add_column_table(tab, tables.Float32Col, "mc_type", 101 * np.ones(len(df)))


def create_final_h5(hfile, hfile_tmp, hfile_tmp2, output_filename):
    """
    Create the final output HDF5 file.
    It copies /instruments and /simulations nodes from the output of hipecta_hdf5_r1_to_dl1.py,
        - and /dl1/event from the created table.
        - and includes the image and pulse_time within the correct path, i.e., /dl1/event/telescope/

    TODO: Define somehow the paths globally so that it can be used .get_node()

    Parameters
    ----------
        hfile: [obj, astropy.table.table.Table] output file
        hfile_tmp: [obj, astropy.table.table.Table] hdf5 file with dl1 parameters. path:
                    `dl1/event/telescope/parameters/LST_LSTCam`
        hfile_tmp2: [obj, astropy.table.table.Table] hdf5 file with images and pulse_time. path:
                    `dl1/event/telescope/image/LST_LSTCam`
        output_filename: [str] name of output file

    Returns
    -------
        None

    """
    # The complevel MUST be set to zero, otherwise this version of libhdf5 does NOT accept the copy_node()
    filter = tables.Filters(
        complevel=0,
        complib="blosc:zstd",
        shuffle=False,
        bitshuffle=False,
        fletcher32=False,
    )

    hfile_out = tables.open_file(output_filename, "w")
    hfile_out.copy_node(
        hfile.root.instrument, newparent=hfile_out.root, recursive=True, filters=filter
    )
    hfile_out.copy_node(
        hfile.root.simulation, newparent=hfile_out.root, recursive=True, filters=filter
    )
    hfile_out.copy_node(hfile_tmp.root.dl1, newparent=hfile_out.root, recursive=True)
    hfile_out.copy_children(
        hfile_tmp2.root.dl1.event.telescope,
        hfile_out.root.dl1.event.telescope,
        recursive=True,
    )

    # Move the telescope table from /instrument/subarray to /instrument (lstchain output file dl1 format)
    hfile_out.move_node(
        "/instrument/subarray/telescope", newparent="/instrument", createparents=True
    )
    hfile_out.close()


def modify_params_table(table, tel_id, focal=28):
    """
    Modify column names and compute missing parameters

    Parameters
    ----------
        table: [obj, astropy.table.table.Table] The table to be modified.
        tel_id: [int] telescope identifier, i.e., 1,2,3,4 for LSTs
        focal: [float] focal length in meters  # TODO: deprecated, done at hipecta/rta level

    Returns
    -------
        None
    """
    # Create the column tel_id

    table.add_column(Column(tel_id * np.ones(len(table)), dtype=int), name="tel_id")

    # Rename `leakage_intensity2` --> `leakage`  # lstchain v0.4.5
    # Rename `leakage_pixel1` --> `leakage1_pixel`          #lstchain v0.5.x
    # Rename `leakage_intensity2` --> `leakage2_intensity`  #lstchain v0.5.x
    table.rename_column("leakage_intensity1", "leakage_intensity_width_1")
    table.rename_column("leakage_intensity2", "leakage_intensity_width_2")
    table.rename_column("leakage_pixel1", "leakage_pixels_width_1")
    table.rename_column("leakage_pixel2", "leakage_pixels_width_2")
    table.rename_column("nb_selected_pixel", "n_pixels")

    # X and Y in meters
    # table['x'] *= focal
    # table['y'] *= focal

    # mc_energy must be computed after merging
    # log of intensity and computation of wl
    table.add_column(np.log10(table["intensity"]), name="log_intensity")
    table.add_column(table["width"] / table["length"], name="wl")


def stack_by_telid(dl1_pointer, focal=28):
    """
    Stack :
        - LST telescopes' parameters into a table
        - Calibrated images and pulse_times into another table

    Parameters
    ----------
        dl1_pointer: [obj, tables.group.Group] pointer of the input hdf5 file `hfile.root.dl1`
        focal: [float] focal length in meters  # TODO: deprecated, done at hipecta/rta level

    Returns
    -------
        Two tables [obj, astropy.table.table.Table] containing the parameters, and the images and pulse_times int their
            respective path
    """

    tels_params = [Table(tel.parameters.read()) for tel in dl1_pointer]
    try:
        tel_ids = [tel["telId"][0] for tel in dl1_pointer]
    except:
        # if the tel_id column does not exist, we assign tel ids by simple iteration
        tel_ids = [i + 1 for i in range(len(tels_params))]

    for tab, tel_id in zip(tels_params, tel_ids):
        modify_params_table(tab, tel_id, focal=focal)

    # tabs = [Table(tel) for tel in tels_params]

    stacked_param = vstack(tels_params)

    images = [Table(tel.calib_pic.read()) for tel in dl1_pointer]

    # adding stupid tel_id to the image table as well
    for image_tab, tel_id in zip(images, tel_ids):
        image_tab.add_column(
            Column(tel_id * np.ones(len(image_tab)), dtype=int), name="tel_id"
        )

    stacked_images = vstack(images)
    if "event_id" not in stacked_images.columns:
        stacked_images.add_column(stacked_param["event_id"])

    try:
        #  HiPeCTA case
        stacked_images.rename_column("eventId", "event_id")
    except KeyError:
        #  HiPeRTA case
        pass

    return stacked_param, stacked_images


def reorganize_dl1(input_filename, output_filename):
    """
    Reorganize the output dl1 files of hiperta/hipecta codes to reach the same structure found in lstchain dl1 files.

    Parameters
    ----------
        input_filename: str
            Input filename
        output_filename: str
            Output filename
    Returns
    -------
        None. It dumps the final hdf5 file with the correct structure.

    """
    hfile = tables.open_file(input_filename, "r")

    # Pointers
    dl1 = hfile.root.dl1
    mc_event = Table(hfile.root.simulation.mc_event.read())

    # Temporal names for temporal files, later erased
    _param = (
        str(os.path.abspath(output_filename).rsplit("/", 1)[0])
        + "/dl1_prams_tmp_"
        + str(os.path.basename(input_filename))
    )
    _images = (
        str(os.path.abspath(output_filename).rsplit("/", 1)[0])
        + "/dl1_imags_tmp_"
        + str(os.path.basename(input_filename))
    )

    # File has not been reorganized yet ! Thus, the path for optics is inside /instrument/subarray/telescope
    # only valid for LSTs !
    focal = hfile.root.instrument.subarray.telescope.optics.col(
        "equivalent_focal_length"
    )[0]
    table_dl1, table_imags = stack_by_telid(dl1, focal=focal)

    # Join together with the mc_events, compute log of mc_energy and dump it
    table_dl1 = join(table_dl1, mc_event, keys="event_id")
    table_dl1.add_column(np.log10(table_dl1["mc_energy"]), name="log_mc_energy")

    # write tmp files
    table_dl1.write(_param, format="hdf5", path=dl1_params_lstcam_key, overwrite=True)
    table_imags.write(
        _images, format="hdf5", path=dl1_images_lstcam_key, overwrite=True
    )

    # open tmp files
    _hfile_param = tables.open_file(_param, "r")
    _hfile_imags = tables.open_file(_images, "r")

    create_final_h5(hfile, _hfile_param, _hfile_imags, output_filename)

    # Add disp_* and mc_type to the parameters table
    add_disp_and_mc_type_to_parameters_table(output_filename, dl1_params_lstcam_key)

    # Close and erase
    _hfile_param.close()
    _hfile_imags.close()
    os.remove(_param)
    os.remove(_images)

    hfile.close()


if __name__ == "__main__":
    args = parser.parse_args()
    reorganize_dl1(args.infile, args.outfile)



#!/usr//bin/env python3

import os
import logging
from pathlib import Path
from lstmcpipe.workflow_management import save_log_to_file


log = logging.getLogger(__name__)


def batch_merge_dl1(
    dict_paths,
    batch_config,
    logs,
    jobid_from_splitting,
    workflow_kind="lstchain",
):
    """
    Function to batch the onsite_mc_merge_and_copy function once the all the r0_to_dl1 jobs (batched by particle type)
    have finished.

    Batch 8 merge_and_copy_dl1 jobs ([train, test] x particle) + the move_dl1 and move_dir jobs (2 per particle).

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings to be passed
        to `merge_dl1` and `compose_batch_command_of_script` functions.
    workflow_kind : str
        Defines workflow kind (lstchain, ctapipe, hiperta)
    logs: dict
        Dictionary with logs files
    jobid_from_splitting: str

    Returns
    -------
    jobids_for_train : str
         Comma-separated str with all the job-ids to be passed to the next
         stage of the workflow (as a slurm dependency)

    """
    log_merge = {}
    all_jobs_merge_stage = []
    debug_log = {}

    log.info("==== START {} ====".format("batch merge_and_copy_dl1_workflow"))
    # TODO Lukas: merging option will come inside the
    #  dict_paths["merge_dl1"]["merging_options"]
#    if isinstance(smart_merge, str):
#        merge_flag = "lst" in smart_merge
#    else:
#        merge_flag = smart_merge
#    log.debug("Merge flag set: {}".format(merge_flag))

    for paths in dict_paths:
        job_logs, jobid_debug = merge_dl1(
            paths["input"],
            paths["output"],
            merging_options=paths.get('options', None),
            batch_configuration=batch_config,
            wait_jobs_split=jobid_from_splitting,
            workflow_kind=workflow_kind,
            slurm_options=paths.get("slurm_options", None),
        )

        log_merge.update(job_logs)
        all_jobs_merge_stage.append(jobid_debug)

    jobids_for_train = ','.join(all_jobs_merge_stage)

    save_log_to_file(log_merge, logs["log_file"], "merge_dl1")
    save_log_to_file(debug_log, logs["debug_file"], workflow_step="merge_dl1")

    log.info("==== END {} ====".format("batch merge_and_copy_dl1_workflow"))

    return jobids_for_train


def merge_dl1(
        input_dir,
        output_file,
        batch_configuration,
        wait_jobs_split="",
        merging_options=None,
        workflow_kind="lstchain",
        slurm_options=None,
):
    """

    Parameters
    ----------
    input_dir: str
    output_file: str
    batch_configuration: dict
    wait_jobs_split: str
    merging_options: dict
    workflow_kind: str
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    log_merge: dict
    jobid_merge: str

    """
    source_environment = batch_configuration["source_environment"]
    slurm_account = batch_configuration["slurm_account"]

    merging_options = "" if merging_options is None else merging_options

    log_merge = {}

    jobo = Path(output_file).parent.joinpath("merging-output.o")
    jobe = Path(output_file).parent.joinpath("merging-error.e")

    cmd = "sbatch --parsable"
    # TODO All slurm options/args can most probable be passed in a more intelligent way
    if slurm_options is not None:
        cmd += f" {slurm_options}"
    else:
        cmd += " -p short"
    if slurm_account != "":
        cmd += f" -A {slurm_account}"
    if wait_jobs_split != "":
        cmd += " --dependency=afterok:" + wait_jobs_split

    cmd += (
        f' -J merge -e {jobe} -o {jobo} --wrap="{source_environment} '
    )

    # command passed changes depending on the workflow_kind
    if workflow_kind == "lstchain":
        cmd += f'lstchain_merge_hdf5_files -d {input_dir} -o {output_file}  {merging_options}'

    elif workflow_kind == "hiperta":
        # HiPeRTA workflow still uses --smart flag (lstchain v0.6.3)
        cmd += (
            f'lstchain_merge_hdf5_files -d {input_dir} -o {output_file}  {merging_options}'
        )
    else:  # ctapipe case
        cmd += f'ctapipe-merge --input-dir {input_dir} --output {output_file}  {merging_options}'

    # IN ALL THE CASES we need to close the " of the wrap
    cmd += '"'

    jobid_merge = os.popen(cmd).read().strip("\n")
    log_merge.update({jobid_merge: cmd})

    log.info(f"Submitted batch job {jobid_merge}")

    return log_merge, jobid_merge



#!/usr/bin/env python

# Code to reduce R0 data to DL1 onsite (La Palma cluster)

import os
import shutil
import logging
from pathlib import Path
from lstmcpipe.workflow_management import save_log_to_file
from lstmcpipe.io.data_management import (
    check_data_path,
)


log = logging.getLogger(__name__)


def batch_process_dl1(
    dict_paths,
    conf_file,
    batch_config,
    logs,
    workflow_kind="lstchain",
    new_production=True,
):
    """
    Batch the dl1 processing jobs by particle type.

    Parameters
    ----------
    dict_paths : dict
        Core dictionary with {stage: PATHS} information
    conf_file : str
        Path to a configuration file. If none is given, a standard configuration is applied
    batch_config : dict
        Dict with source environment (to select the desired conda environment to run the r0/1_to_dl1 stage),
        and the slurm user account.
    workflow_kind: str
        One of the supported pipelines. Defines the command to be run on r0 files
    new_production: bool
        Whether to analysis simtel or reprocess existing dl1 files.
    logs: dict
        Dictionary con logs files

    Returns
    -------
    jobids_dl1_processing_stage : str
        string, separated by commas, containing all the jobids of this stage
    """
    log_process_dl1 = {}
    debug_log = {}
    jobids_dl1_processing_stage = []

    log.info("==== START {} dl1 processing ====".format(workflow_kind))

    if new_production:

        for paths in dict_paths["r0_to_dl1"]:
            job_logs, jobid = r0_to_dl1(
                paths["input"],
                paths["output"],
                config_file=conf_file,
                batch_config=batch_config,
                workflow_kind=workflow_kind,
                slurm_options=paths.get("slurm_options", None),
            )
            log_process_dl1.update(job_logs)
            jobids_dl1_processing_stage.append(jobid)
            debug_log[jobid] = f'r0_dl1 job from input dir: {paths["input"]}'

    else:

        for paths in dict_paths["dl1ab"]:

            job_logs, jobid = reprocess_dl1(
                paths["input"],
                paths["output"],
                config_file=conf_file,
                batch_config=batch_config,
                workflow_kind=workflow_kind,
                slurm_options=paths.get("slurm_options", None),
            )

            log_process_dl1.update(job_logs)
            jobids_dl1_processing_stage.append(jobid)
            debug_log[jobid] = f'dl1ab job from input dir: {paths["input"]}'

    # Create a string to be directly passed
    jobids_dl1_processing_stage = ",".join(jobids_dl1_processing_stage)

    save_log_to_file(log_process_dl1, logs["log_file"], "r0_to_dl1")
    save_log_to_file(debug_log, logs["debug_file"], workflow_step="r0_to_dl1")

    log.info("==== END {} dl1 processing ====".format(workflow_kind))

    return jobids_dl1_processing_stage


def r0_to_dl1(
    input_dir,
    output_dir,
    workflow_kind="lstchain",
    config_file=None,
    batch_config=None,
    debug_mode=False,
    keep_rta_file=False,
    slurm_options=None,
):
    """
    R0 to DL1 MC onsite conversion.
    Organizes files and launches slurm jobs in two slurm arrays.


    Parameters
    ----------
    input_dir : str or Path
        path to the files directory to analyse
    output_dir : str or Path
        Output path to store files
    config_file : str or Path
        Path to the {lstchain, ctapipe, HiPeRTA} configuration file
    batch_config : dict
        Dictionary containing the full (source + env) source_environment and the slurm_account strings.
        ! NOTE : train_pipe AND dl1_to_dl2 **MUST** be run with the same environment.
    workflow_kind: str
        One of the supported pipelines. Defines the command to be run on r0 files
    slurm_options: str
        Extra slurm options to be passed

    # HIPERTA ARGUMENTS
    keep_rta_file : bool
        Flag to indicate whether to keep (keep_rta_file = True) or remove (keep_rta_file = Flase ) the
        `dl1v06_reorganized_*.h5` output file (hiperta_r0_dl1 and re-organiser stages).
        Argument to be passed to the hiperta_r0_to_dl1lstchain script.
    debug_mode : bool
        Flag to activate debug_mode. Only compatible with `hiperta` workflow kind, i.e.,
        HiPeRTA functionality. DEFAULT=False.

    Returns
    -------
    jobid2log : dict
        dictionary log containing {jobid: batch_cmd} information
    jobids_r0_dl1
        A list of all the jobs sent for input dir
    """

    log.info("Starting R0 to DL1 processing for files in dir : {}".format(input_dir))

    source_environment = batch_config["source_environment"]
    slurm_account = batch_config["slurm_account"]

    if workflow_kind == "lstchain":
        base_cmd = f"{source_environment} lstmcpipe_lst_core_r0_dl1 -c {config_file} "
        jobtype_id = "LST"
    elif workflow_kind == "ctapipe":
        base_cmd = f"{source_environment} lstmcpipe_cta_core_r0_dl1 -c {config_file} "
        jobtype_id = "CTA"
    elif workflow_kind == "hiperta":
        base_cmd = (
            f"{source_environment} lstmcpipe_rta_core_r0_dl1 -c {config_file} "
        )
        if keep_rta_file:
            base_cmd += " --keep_rta_file"
        if debug_mode:
            base_cmd += " --debug_mode"

        jobtype_id = "RTA"
    else:
        base_cmd = ''
        jobtype_id = ''
        log.critical("Please, select an allowed workflow kind.")
        exit(-1)

    log.info("Working on DL0 files in {}".format(input_dir))

    check_data_path(input_dir)
    raw_files_list = [file.resolve().as_posix() for file in Path(input_dir).rglob("**/*.simtel.gz")]

    if len(raw_files_list) < 50:
        dl1_files_per_job = 20
    else:
        dl1_files_per_job = 50

    with open("r0_to_dl1.list", "w+") as newfile:
        for f in raw_files_list:
            newfile.write(f)
            newfile.write("\n")

    log.info("{} raw R0 files".format(len(raw_files_list)))

    # If file exists, means that prod is being re-run
    output_dir = Path(output_dir)
    if output_dir.exists() and any(output_dir.iterdir()):
        shutil.rmtree(output_dir)

    job_logs_dir = output_dir.joinpath("job_logs_r0dl1")
    Path(job_logs_dir).mkdir(exist_ok=True, parents=True)

    log.info("DL1 DATA DIR: {}".format(output_dir))

    jobid2log, jobids_r0_dl1 = submit_dl1_jobs(
        input_dir,
        output_dir,
        base_cmd=base_cmd,
        file_list=raw_files_list,
        job_type_id=jobtype_id,
        dl1_files_per_batched_job=dl1_files_per_job,
        job_logs_dir=job_logs_dir,
        slurm_account=slurm_account,
        filelist_name="r0_to_dl1",
        slurm_options=slurm_options,
    )

    # copy config into working dir
    if config_file is not None:
        shutil.copyfile(
            config_file,
            job_logs_dir.joinpath(Path(config_file).name)
        )

    # save file lists into logs
    shutil.move("r0_to_dl1.list", job_logs_dir)

    # return it log dictionary
    return jobid2log, jobids_r0_dl1


def reprocess_dl1(
    input_dir,
    output_dir,
    workflow_kind="lstchain",
    config_file=None,
    batch_config=None,
    dl1_files_per_job=50,
    slurm_options=None,
):
    """
    Reprocessing of existing dl1 files.
    Organizes files and launches slurm jobs in two slurm arrays.
    The same train/test split performed with the earlier r0 to dl1
    processing is used.

    Parameters
    ----------
    input_dir : str
        path to the files directory to analyse
    output_dir : str or Path
        Output path to store files
    config_file :str
        Path to a configuration file. If none is given, the standard configuration of the selected pipeline is applied
    batch_config : dict
        Dictionary containing the (full) source_environment and the slurm_account strings.
        ! NOTE : train_pipe AND dl1_to_dl2 **MUST** be run with the same environment.
    workflow_kind: str
        One of the supported pipelines. Defines the command to be run on r0 files
    dl1_files_per_job: int
        Number of dl1 files to be processed per job array that was batched.
    slurm_options: str
        Extra slurm options to be passed to the sbatch command

    Returns
    -------
    jobid2log : dict
        dictionary log containing {jobid: batch_cmd} information
    jobids_dl1_dl1
        A list of all the jobs sent for input dir
    """
    log.info("Applying DL1ab on DL1 files in {}".format(input_dir))

    source_environment = batch_config["source_environment"]
    slurm_account = batch_config["slurm_account"]

    if workflow_kind == "lstchain":
        base_cmd = f"{source_environment} lstmcpipe_lst_core_dl1ab -c {config_file} "
        jobtype_id = "LST"
    elif workflow_kind == "ctapipe":
        base_cmd = f"{source_environment} lstmcpipe_cta_core_r0_dl1 -c {config_file} "
        jobtype_id = "CTA"
    else:
        base_cmd = ""
        jobtype_id = ""
        log.critical("Please, selected an allowed workflow kind.")
        exit(-1)

    log.info("Working on DL1 files in {}".format(input_dir))

    check_data_path(input_dir)
    dl1ab_filelist = [file.resolve().as_posix() for file in Path(input_dir).rglob("*.h5")]

    log.info("{} DL1 files".format(len(dl1ab_filelist)))

    with open("dl1ab.list", "w+") as newfile:
        for f in dl1ab_filelist:
            newfile.write(f)
            newfile.write("\n")

    job_logs_dir = output_dir.joinpath("job_logs_dl1ab")
    Path(job_logs_dir).mkdir(exist_ok=True)

    log.info("DL1ab DATA DIR: {}".format(output_dir))

    jobid2log, jobids_dl1_dl1 = submit_dl1_jobs(
        input_dir,
        output_dir,
        base_cmd=base_cmd,
        file_list=dl1ab_filelist,
        job_type_id=jobtype_id,
        dl1_files_per_batched_job=dl1_files_per_job,
        job_logs_dir=job_logs_dir,
        slurm_account=slurm_account,
        filelist_name="dl1ab",
        slurm_options=slurm_options,
    )

    # copy config into working dir
    if config_file is not None:
        shutil.copyfile(
            config_file,
            job_logs_dir.joinpath(Path(config_file).name)
        )

    # save file lists into logs
    shutil.move("dl1ab.list", job_logs_dir)

    # return it log dictionary
    return jobid2log, jobids_dl1_dl1


def submit_dl1_jobs(
    input_dir,
    output_dir,
    base_cmd,
    file_list,
    job_type_id,
    dl1_files_per_batched_job,
    job_logs_dir,
    slurm_account,
    filelist_name,
    n_jobs_parallel=100,
    dl1_processing="r0_dl1",
    slurm_options=None,
):
    """
    Compose sbatch command and batches it

    Parameters
    ----------
    input_dir: str
    output_dir: str
    base_cmd: str
    file_list: list
    job_type_id: str
    dl1_files_per_batched_job: int
    job_logs_dir: Path
    slurm_account: str
    filelist_name: str
    n_jobs_parallel: int
    dl1_processing: str
    slurm_options: str
        Extra slurm options to be passed

    Returns
    -------
    jobid2log: dict
    jobid: str

    """

    jobid2log = {}

    log.info("output dir: {}".format(output_dir))

    number_of_sublists = len(file_list) // dl1_files_per_batched_job + int(
        len(file_list) % dl1_files_per_batched_job > 0
    )

    for i in range(number_of_sublists):
        output_file = job_logs_dir.joinpath("{}_{}.sublist".format(filelist_name, i)).resolve().as_posix()
        with open(output_file, "w+") as out:
            for line in file_list[
                i * dl1_files_per_batched_job: dl1_files_per_batched_job * (i + 1)
            ]:
                out.write(line)
                out.write("\n")

    log.info(f"{number_of_sublists} files generated for list of files at {input_dir}")

    sublist_names = [f.as_posix() for f in Path(job_logs_dir).glob("*.sublist")]  # Number of sublists ??

    default_slurm_options = {
        "output": job_logs_dir.joinpath("job_%A_%a.o").as_posix(),
        "error": job_logs_dir.joinpath("job_%A_%a.e").as_posix(),
        "array": f"0-{len(sublist_names)-1}%{n_jobs_parallel}",
        "job-name": f"{job_type_id}-{dl1_processing}"
    }

    # All files to long queue
    default_slurm_options.update({"partition": "long"})
    # `sbatch -A` is the --account slurm argument
    if slurm_account != '':
        default_slurm_options.update({"account": slurm_account})

    # start 1 jobarray with all files included. The job selects its file based on its task id
    cmd = f'{base_cmd} -f {" ".join(sublist_names)} --output_dir {output_dir}'
    slurm_cmd = "sbatch --parsable"
    for key, value in default_slurm_options.items():
        slurm_cmd += f" --{key} {value}"
    if slurm_options is not None:
        slurm_cmd += f" {slurm_options}"
    slurm_cmd += f' --wrap="{cmd}"'

    jobid = os.popen(slurm_cmd).read().strip("\n")
    log.debug(f"Submitted batch job {jobid}")

    jobid2log.update({jobid: slurm_cmd})

    return jobid2log, jobid



import json
import subprocess
import logging


log = logging.getLogger(__name__)


def create_dl1ab_tuned_config(
    base_config_path, config_output_path, observed_data_path, mc_path
):
    """
    Create a new lstchain config with noise parameters added.
    If there are noise parameters in the input config,
    this will fail with an Exception.

    Parameters:
    -----------
    base_config_path: str or Path-like object
        Input config. Will not be overwritten.
    config_output_path: str or Path-like object
        Where to save the new config file.
    observed_data_path: str or Path-like object
        Path to a dl1 file with observed data from which to
        calculate the parameters.
    observed_mc_path: str or Path-like object
        Path to a dl1 file with simulated data from which to
        calculate the parameters.

    Return:
    -------
    config_output_path
    """
    with open(base_config_path) as f:
        base = json.load(f)

    log.info("Running lstchain_tune_nsb to get the noise tuning parameters")
    log.info(f"Comparing mc file {mc_path} and observed file {observed_data_path}")
    cmd = [
        "lstchain_tune_nsb",
        f"--config={base_config_path}",
        f"--input-mc={mc_path}",
        f"--input-data={observed_data_path}",
    ]
    res = subprocess.run(cmd, capture_output=True)
    # Log output is directed to stderr not stdout
    log_output = res.stderr.decode()

    # The script dumps the dict as the very last output
    dict_string = "{" + log_output.rsplit("{", -1)[-1].strip()

    # This should now load into json
    parameters = json.loads(dict_string)

    modifier_settings = base.get("image_modifier", {})
    for key, value in parameters.items():
        if key in modifier_settings:
            log.warning(
                f"Overwriting image modifier parameter {key} from the base config with "
                f"the new value {value}."
            )
        modifier_settings[key] = value
    base["image_modifier"] = modifier_settings

    with open(config_output_path, "w") as f:
        json.dump(base, f, indent=2)

    return config_output_path



#!/usr/bin/env python

# T. Vuillaume & E. Garcia - 2021

from pathlib import Path
import argparse
from astropy.table import QTable
import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
from pyirf.utils import cone_solid_angle
import ctaplot
from astropy.visualization import quantity_support


def plot_summary_from_file(filename, axes=None, **kwargs):

    if axes is None:
        fig, axes = plt.subplots(2, 2, figsize=(15, 15))

    sens_table = read_sensitivity_table(filename)
    plot_sensitivity_from_table(sens_table, ax=axes.ravel()[0], **kwargs)

    plot_angular_resolution_from_file(filename, ax=axes.ravel()[1], **kwargs)

    plot_energy_resolution_from_file(filename, ax=axes.ravel()[2], **kwargs)

    plot_effective_area_from_file(filename, ax=axes.ravel()[3], **kwargs)

    axes.ravel()[0].get_figure().tight_layout()

    return axes


def read_sensitivity_table(irf_file):
    """ """
    sensitivity = QTable.read(irf_file, hdu="SENSITIVITY")[1:-1]
    return sensitivity


def plot_gh_cut_per_energy(filename, ax=None, **kwargs):
    """ """
    ax = plt.gca() if ax is None else ax

    gh_cut = QTable.read(filename, hdu="GH_CUTS")[1:-1]

    kwargs.setdefault("ls", "")
    ax.errorbar(
        0.5 * (gh_cut["low"] + gh_cut["high"]).to_value(u.TeV),
        gh_cut["cut"],
        xerr=0.5 * (gh_cut["high"] - gh_cut["low"]).to_value(u.TeV),
        **kwargs,
    )

    ax.legend()
    ax.set_ylabel("G/H-cut")
    ax.set_xlabel(r"$E_\mathrm{reco} / \mathrm{TeV}$")
    ax.set_xscale("log")

    return ax


def plot_sensitivity_from_table(sens_table, ax=None, **kwargs):
    """ """

    ax = plt.gca() if ax is None else ax

    sens_unit = u.Unit("erg cm-2 s-1")

    e = sens_table["reco_energy_center"]
    s = e ** 2 * sens_table["flux_sensitivity"]
    w = sens_table["reco_energy_high"] - sens_table["reco_energy_low"]

    ax.errorbar(
        e.to_value(u.TeV),
        s.to_value(sens_unit),
        xerr=w.to_value(u.TeV) / 2,
        **kwargs,
    )

    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.grid(True, which="both")

    # Style settings
    ax.set_title("Minimal Flux Needed for 5σ Detection in 50 hours")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_ylabel(
        rf"$(E^2 \cdot \mathrm{{Flux Sensitivity}}) /$ ({sens_unit.to_string('latex')})"
    )

    return ax


def plot_sensitivity_from_file(irf_file, ax=None, **kwargs):

    tab = read_sensitivity_table(irf_file)
    ax = plot_sensitivity_from_table(tab, ax=ax, **kwargs)
    return ax


def plot_magic_2014(ax=None, **kwargs):
    ax = plt.gca() if ax is None else ax
    file = "/fefs/home/thomas.vuillaume/software/cta-observatory/cta-lstchain/lstchain/spectra/data/magic_sensitivity.txt"
    magic_2014 = np.loadtxt(file, skiprows=1)

    kwargs.setdefault("label", "MAGIC (Aleksic et al. 2014)")
    ax.loglog(
        magic_2014[:, 0] / 1e3,
        magic_2014[:, 3] * np.power(magic_2014[:, 0] / 1e3, 2),
        **kwargs,
    )

    return ax


def plot_effective_area_from_file(file, all_cuts=False, ax=None, **kwargs):
    """ """

    ax = plt.gca() if ax is None else ax

    if all_cuts:
        names = ["", "_NO_CUTS", "_ONLY_GH", "_ONLY_THETA"]
    else:
        names = tuple([""])

    label_basename = kwargs["label"] if "label" in kwargs else ""

    kwargs.setdefault("ls", "")

    for name in names:

        area = QTable.read(file, hdu="EFFECTIVE_AREA" + name)[0]

        kwargs["label"] = label_basename + name.replace("_", " ")

        ax.errorbar(
            0.5 * (area["ENERG_LO"] + area["ENERG_HI"]).to_value(u.TeV)[1:-1],
            area["EFFAREA"].to_value(u.m ** 2).T[1:-1, 0],
            xerr=0.5 * (area["ENERG_LO"] - area["ENERG_HI"]).to_value(u.TeV)[1:-1],
            **kwargs,
        )

        # Style settings
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel("True energy / TeV")
        ax.set_ylabel("Effective collection area / m²")
        ax.grid(which="both")
        ax.legend()
        ax.grid(True, which="both")

    return ax


def plot_psf_from_file(filename):

    psf_table = QTable.read(filename, hdu="PSF")[0]
    # select the only fov offset bin
    psf = psf_table["RPSF"].T[:, 0, :].to_value(1 / u.sr)

    offset_bins = np.append(psf_table["RAD_LO"], psf_table["RAD_HI"][-1])
    phi_bins = np.linspace(0, 2 * np.pi, 100)

    # Let's make a nice 2d representation of the radially symmetric PSF
    r, phi = np.meshgrid(offset_bins.to_value(u.deg), phi_bins)

    # look at a single energy bin
    # repeat values for each phi bin
    center = 0.5 * (psf_table["ENERG_LO"] + psf_table["ENERG_HI"])

    fig = plt.figure(figsize=(15, 5))
    axs = [fig.add_subplot(1, 3, i, projection="polar") for i in range(1, 4)]

    for bin_id, ax in zip([10, 20, 30], axs):
        image = np.tile(psf[bin_id], (len(phi_bins) - 1, 1))

        ax.set_title(f"PSF @ {center[bin_id]:.2f} TeV")
        ax.pcolormesh(phi, r, image)
        ax.set_ylim(0, 0.25)
        ax.set_aspect(1)

    fig.tight_layout()

    return axs


def plot_theta_cut_from_file(filename, ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    rad_max = QTable.read(filename, hdu="RAD_MAX")[0]

    kwargs.setdefault("ls", "")
    ax.errorbar(
        0.5 * (rad_max["ENERG_LO"] + rad_max["ENERG_HI"])[1:-1].to_value(u.TeV),
        rad_max["RAD_MAX"].T[1:-1, 0].to_value(u.deg),
        xerr=0.5 * (rad_max["ENERG_HI"] - rad_max["ENERG_LO"])[1:-1].to_value(u.TeV),
        **kwargs,
    )

    ax.legend()
    ax.set_ylabel("θ-cut / deg²")
    ax.set_xlabel(r"$E_\mathrm{reco} / \mathrm{TeV}$")
    ax.set_xscale("log")

    return ax


def plot_angular_resolution_from_file(filename, ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    ang_res = QTable.read(filename, hdu="ANGULAR_RESOLUTION")[1:-1]

    kwargs.setdefault("ls", "")
    ax.errorbar(
        0.5
        * (ang_res["true_energy_low"] + ang_res["true_energy_high"]).to_value(u.TeV),
        ang_res["angular_resolution"].to_value(u.deg),
        xerr=0.5
        * (ang_res["true_energy_high"] - ang_res["true_energy_low"]).to_value(u.TeV),
        **kwargs,
    )

    # Style settings
    #     ax.set_xlim(1.e-2, 2.e2)
    #     ax.set_ylim(2.e-2, 1)
    ax.set_xscale("log")
    ax.set_xlabel("True energy / TeV")
    ax.set_ylabel("Angular Resolution / deg")
    ax.grid(True, which="both")
    ax.legend(loc="best")

    return ax


def plot_energy_dispersion_from_file(filename):

    edisp = QTable.read(filename, hdu="ENERGY_DISPERSION")[0]

    e_bins = edisp["ENERG_LO"][1:]
    migra_bins = edisp["MIGRA_LO"][1:]

    fig = plt.gca()
    plt.pcolormesh(
        e_bins.to_value(u.TeV),
        migra_bins,
        edisp["MATRIX"].T[1:-1, 1:-1, 0].T,
        cmap="inferno",
    )

    plt.xscale("log")
    plt.yscale("log")
    plt.colorbar(label="PDF Value")

    plt.xlabel(r"$E_\mathrm{True} / \mathrm{TeV}$")
    plt.ylabel(r"$E_\mathrm{Reco} / E_\mathrm{True}$")

    return fig


def plot_energy_resolution_from_file(filename, ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    bias_resolution = QTable.read(filename, hdu="ENERGY_BIAS_RESOLUTION")[1:-1]

    kwargs.setdefault("ls", "")

    # Plot function
    ax.errorbar(
        0.5
        * (
            bias_resolution["true_energy_low"] + bias_resolution["true_energy_high"]
        ).to_value(u.TeV),
        bias_resolution["resolution"],
        xerr=0.5
        * (
            bias_resolution["true_energy_high"] - bias_resolution["true_energy_low"]
        ).to_value(u.TeV),
        **kwargs,
    )
    ax.set_xscale("log")

    # Style settings
    ax.set_xlabel(r"$E_\mathrm{True} / \mathrm{TeV}$")
    ax.set_ylabel("Energy resolution")
    ax.grid(True, which="both")
    ax.legend()

    return ax


def plot_background_rate_from_file(filename, ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    rad_max = QTable.read(filename, hdu="RAD_MAX")[0]
    bg_rate = QTable.read(filename, hdu="BACKGROUND")[0]

    reco_bins = np.append(bg_rate["ENERG_LO"], bg_rate["ENERG_HI"][-1])

    # first fov bin, [0, 1] deg
    fov_bin = 0
    rate_bin = bg_rate["BKG"].T[:, fov_bin]

    # interpolate theta cut for given e reco bin
    e_center_bg = 0.5 * (bg_rate["ENERG_LO"] + bg_rate["ENERG_HI"])
    e_center_theta = 0.5 * (rad_max["ENERG_LO"] + rad_max["ENERG_HI"])
    theta_cut = np.interp(e_center_bg, e_center_theta, rad_max["RAD_MAX"].T[:, 0])

    # undo normalization
    rate_bin *= cone_solid_angle(theta_cut)
    rate_bin *= np.diff(reco_bins)

    if "ls" not in kwargs and "linestyle" not in kwargs:
        kwargs["ls"] = ""

    ax.errorbar(
        e_center_bg.to_value(u.TeV)[1:-1],
        rate_bin.to_value(1 / u.s)[1:-1],
        xerr=np.diff(reco_bins).to_value(u.TeV)[1:-1] / 2,
        **kwargs,
    )

    # Style settings
    ax.set_xscale("log")
    ax.set_xlabel(r"$E_\mathrm{Reco} [\mathrm{TeV}]$")
    ax.set_ylabel("Background rate [Hz]")
    ax.grid(True, which="both")
    ax.legend()
    ax.set_yscale("log")

    return ax


def plot_magic_bkg_rate(ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    table = ctaplot.ana.get_magic_sensitivity()

    emin = table["e_min"].to_value(u.TeV)
    emax = table["e_max"].to_value(u.TeV)
    ecenter = table["e_center"].to_value(u.TeV)
    bkg_rate = table["background_rate"].to_value(1 / u.s)
    bkg_rate_err = table["background_rate_err"].to_value(1 / u.s)

    kwargs.setdefault("label", "MAGIC (Aleksić et al, 2016)")

    ax.errorbar(
        ecenter,
        bkg_rate,
        xerr=[ecenter - emin, emax - ecenter],
        yerr=bkg_rate_err,
        **kwargs,
    )

    ax.set_xlabel(r"$E_\mathrm{Reco} [\mathrm{TeV}]$")
    ax.set_ylabel("Background rate [Hz]")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.grid(True, which="both")
    ax.legend()

    return ax


def plot_energy_bias_from_file(filename, ax=None, **kwargs):

    ax = plt.gca() if ax is None else ax

    bias_resolution = QTable.read(filename, hdu="ENERGY_BIAS_RESOLUTION")[1:-1]

    kwargs.setdefault("ls", "")

    # Plot function
    ax.errorbar(
        0.5
        * (
            bias_resolution["true_energy_low"] + bias_resolution["true_energy_high"]
        ).to_value(u.TeV),
        bias_resolution["bias"],
        xerr=0.5
        * (
            bias_resolution["true_energy_high"] - bias_resolution["true_energy_low"]
        ).to_value(u.TeV),
        **kwargs,
    )
    ax.set_xscale("log")

    # Style settings
    ax.set_xlabel(r"$E_\mathrm{True} / \mathrm{TeV}$")
    ax.set_ylabel("Energy bias")
    ax.grid(True, which="both")
    ax.legend()

    return ax


def plot_sensitivity_ratio(sensitivity_tables, baseline_index=0, ax=None, labels=None, **kwargs):
    """
    Plot the ratio of sensitivities as a function of the energy

    Parameters
    ----------
    sensitivity_tables: list
        list of sensitivity tables
    baseline_index: int
        index of the baseline to use in the list
    ax: pyplot.axis
    labels: list 
        list of labels to use
    kwargs: kwargs for the plot

    Returns
    -------
    ax: pyplot.axis
    """

    ax = plt.gca() if ax is None else ax 
    
    sens_table_baseline = sensitivity_tables[baseline_index]
    e = sens_table_baseline["reco_energy_center"]
    w = sens_table_baseline["reco_energy_high"] - sens_table_baseline["reco_energy_low"]
    s_baseline = e ** 2 * sens_table_baseline["flux_sensitivity"]
    
    for ii, sens_table in enumerate(sensitivity_tables):
        s = e ** 2 * sens_table["flux_sensitivity"]
        if labels is not None:
            kwargs['label'] = labels[ii]
        with quantity_support():
            ax.errorbar(
                e,
                s/s_baseline,
                xerr=w / 2,
                **kwargs,
            )
    ax.set_title('Sensitivity ratio (lower is better)')
    ax.set_xscale('log')
    ax.grid(True, which='both')
    ax.set_xlabel(f'Energy / {e.unit}')
        
    return ax


def plot_sensitivity_ratio_from_files(filelist, baseline_index=0, ax=None, **kwargs):
    """
    Plot the ratio of sensitivities as a function of the energy

    Parameters
    ----------
    sensitivity_tables: list
        list of sensitivity tables
    baseline_index: int
        index of the baseline to use in the list
    ax: pyplot.axis
    kwargs: kwargs for the plot

    Returns
    -------
    ax: pyplot.axis
    """

    sens_tables = [read_sensitivity_table(file) for file in filelist]
    labels = [Path(file).name for file in filelist]
    ax = plot_sensitivity_ratio(sens_tables, baseline_index=baseline_index, ax=ax, labels=labels, **kwargs)
    return ax


def main():

    parser = argparse.ArgumentParser(
        description="Produce lstMCpipe IRFs plot from a sensitivity.fits.gz file"
    )

    # Required arguments
    parser.add_argument(
        "--filename", "-f", type=str, dest="filename", help="Input filename"
    )
    parser.add_argument(
        "--outfile", "-o", type=Path, dest="outfile", help="Output filename"
    )

    args = parser.parse_args()

    plot_summary_from_file(args.filename)

    args.outfile.parent.mkdir(exist_ok=True)
    plt.savefig(args.outfile)


if __name__ == "__main__":
    main()
